import inspect
import sys
import importlib
import ntpath
import importlib
from collections import defaultdict
from typing import Dict, Any, Optional
from pathlib import Path


class Generator:

    def __init__(self, generator_types: Dict[str, Any], known_module_paths: Dict[str, str]={}) -> None:
        self.generator_types = generator_types
        self.known_module_paths = known_module_paths
        self.global_imports = {}
        self.local_imports = defaultdict(list)
        self.locals = []

        self.serialized_global_imports = []
        self.serialized_local_imports = []
        self.serialized_locals = []

    def gather_required_items(self, generator_type: str) -> Dict[str, Any]:
        generator_item = self.generator_types.get(generator_type)
        path = inspect.getfile(generator_item)

        package_dir = Path(path).resolve().parent
        package_dir_path = str(package_dir)
        package_dir_module = package_dir_path.split('/')[-1]

        package = ntpath.basename(path)
        package_slug = package.split('.')[0]
        spec = importlib.util.spec_from_file_location(f'{package_dir_module}.{package_slug}', path)

        if path not in sys.path:
            sys.path.append(str(package_dir.parent))

        module = importlib.util.module_from_spec(spec)
        sys.modules[module.__name__] = module

        spec.loader.exec_module(module)

        local_file_items = [item for item in dir(module) if not item.startswith("__")]
        modules = {
            member_name: member for member_name, member in inspect.getmembers(module) if member_name in local_file_items
        }

        for module_name, module in modules.items():

            try:
                same_source_file = inspect.getfile(module) == inspect.getfile(generator_item)

                if same_source_file and module_name != generator_item.__name__:
                    self.locals.append(module)

            except TypeError:
                pass

        self.locals.append(generator_item)

        return modules

    def collect_imports(self, generator_type: Optional[str], modules: Dict[str, Any]):

        generator_item = self.generator_types.get(generator_type)
        generator_item_name = None


        if generator_item:
            generator_item_name = generator_item.__name__

        for module_name, module in modules.items():
            if inspect.ismodule(module):
                self.global_imports[module.__name__] = module.__name__

            elif module_name != generator_item_name and module not in self.locals:
                
                module_path = self.known_module_paths.get(module_name, module.__module__)
                minimum_viable_import_path = module_path.split('.')


                run_minimum_path_search = True
                while run_minimum_path_search:
                    minimum_viable_import_path.pop()
                    
                    if len(minimum_viable_import_path) > 0:
                        candidate_module_path = '.'.join(minimum_viable_import_path)

                        source_module = importlib.import_module(candidate_module_path)
                        source_module_members = {
                            source_module_name: source_module for source_module_name, source_module in inspect.getmembers(source_module)
                        }

                        if module_name in source_module_members:
                            module_path = candidate_module_path
                    else:
                        run_minimum_path_search = False
                    
                self.local_imports[module_path].append(module_name)

    def serialize_items(self):

        self.serialized_global_imports = [
            f'import {global_import}' for global_import in self.global_imports
        ]

        for local_import_module, local_import_items in self.local_imports.items():
            serialized_import_items = '\n\t'.join([
                f'{local_import_item},' for local_import_item in local_import_items
            ])

            self.serialized_local_imports.append(
                f'from {local_import_module} import (\n\t{serialized_import_items}\n)'
            )

        self.serialized_locals = [
            inspect.getsource(local) for local in self.locals
        ]

    



import uuid
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from .influxdb_config import InfluxDBConfig


try:
    from influxdb_client import Point
    from influxdb_client.client.influxdb_client_async import InfluxDBClientAsync
    has_connector = True

except Exception:
    Point = None
    InfluxDBClientAsync = None
    has_connector = False


class InfluxDB:

    def __init__(self, config: InfluxDBConfig) -> None:
        self.host = config.host
        self.token = config.token
        self.protocol = 'https' if config.secure else 'http'
        self.organization = config.organization
        self.connect_timeout = config.connect_timeout

        self.events_bucket_name = config.events_bucket
        self.metrics_bucket_name = config.metrics_bucket
        self.streams_bucket_name = config.streams_bucket
        self.shared_metrics_bucket_name = f'{config.metrics_bucket}_shared'

        self.experiments_bucket_name = config.experiments_bucket
        self.variants_bucket_name = f'{config.experiments_bucket}_variants'
        self.mutations_bucket_name = f'{config.experiments_bucket}_mutations'

        self.errors_bucket_name = f'{config.metrics_bucket}_errors'
        self.custom_bucket_name = f'{config.metrics_bucket}_custom'

        self.session_system_metrics_bucket_name = f'{config.system_metrics_bucket}_session'
        self.stage_system_metrics_bucket_name = f'{config.system_metrics_bucket}_stage'

        self.client = None
        self.write_api = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to InfluxDB at - {self.protocol}://{self.host} - for Organization - {self.organization}')

        self.client = InfluxDBClientAsync(
            f'{self.protocol}://{self.host}',
            token=self.token,
            org=self.organization,
            timeout=self.connect_timeout
        )

        self.write_api = self.client.write_api()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to InfluxDB at - {self.protocol}://{self.host} - for Organization - {self.organization}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to Bucket - {self.session_system_metrics_bucket_name}')

        metrics_sets: List[SessionMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.system_metrics_set_id}')
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics)

        points = []
        for metrics_set in metrics_sets:

            point = Point(metrics_set.name)
            tags = [
                ("name", metrics_set.name),
                ("group", metrics_set.group)
            ]

            for tag_name, tag_value in tags:
                point.tag(tag_name, tag_value)

            for field, value in metrics_set.record.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.name}:{metrics_set.group}:{field}')
                point.field(field, value)

            points.append(point)

        await self.write_api.write(
            bucket=self.session_system_metrics_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to Bucket - {self.session_system_metrics_bucket_name}')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to Bucket - {self.stage_system_metrics_bucket_name}')

        metrics_sets: List[SystemMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.system_metrics_set_id}')
            
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics)

        points = []
        for metrics_set in metrics_sets:

            point = Point(metrics_set.name)
            tags = [
                ("name", metrics_set.name),
                ("group", metrics_set.group)
            ]

            for tag_name, tag_value in tags:
                point.tag(tag_name, tag_value)

            for field, value in metrics_set.record.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.name}:{metrics_set.group}:{field}')
                point.field(field, value)

            points.append(point)

        await self.write_api.write(
            bucket=self.stage_system_metrics_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metrics to Bucket - {self.stage_system_metrics_bucket_name}')

    async def submit_streams(self, stage_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Bucket - {self.streams_bucket_name}')

        points = []
        for stage_name, stream in stage_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams - {stage_metrics}:{stream.stream_set_id}')
            
            stream_name = f'{stage_name}_streams'

            for group_name, group in stream.grouped.items():
                point = Point(stream_name)
                tags = [
                    ("name", stream_name),
                    ("stage", stage_name),
                    ("group", group_name)
                ]

                for tag_name, tag_value in tags:
                    point.tag(tag_name, tag_value)

                for field, value in group.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stream - {stage_name}:{group_name}:{field}')
                    point.field(field, value)

                points.append(point)

        await self.write_api.write(
            bucket=self.streams_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to Bucket - {self.streams_bucket_name}')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Bucket - {self.experiments_bucket_name}')

        points = []
        for experiment in experiment_metrics.experiment_summaries:
            point = Point(experiment.experiment_name)


            for tag in experiment.tags:
                point.tag(tag.name, tag.value)

            for field, value in experiment.stats.items():
                point.field(f'{experiment.experiment_name}_{field}', value)

            points.append(point)

        await self.write_api.write(
            bucket=self.experiments_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to Bucket - {self.experiments_bucket_name}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Bucket - {self.variants_bucket_name}')

        points = []
        for variant in experiment_metrics.variant_summaries:
            point = Point(variant.variant_name)


            for tag in variant.tags:
                point.tag(tag.name, tag.value)

            for field, value in variant.stats.items():
                point.field(f'{variant.variant_name}_{field}', value)

            points.append(point)

        await self.write_api.write(
            bucket=self.variants_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to Bucket - {self.variants_bucket_name}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to Bucket - {self.mutations_bucket_name}')

        points = []
        for mutation in experiment_metrics.mutation_summaries:
            point = Point(mutation.mutation_name)


            for tag in mutation.tags:
                point.tag(tag.name, tag.value)

            for field, value in mutation.stats.items():
                point.field(f'{mutation.mutation_name}_{field}', value)

            points.append(point)

        await self.write_api.write(
            bucket=self.mutations_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to Bucket - {self.mutations_bucket_name}')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Bucket - {self.events_bucket_name}')

        points = []
        for event in events:
            point = Point(event.name)


            for tag in event.tags:
                point.tag(tag.name, tag.value)

            point.field(f'{event.name}_time', event.time)

            if event.success:
                point.field(f'{event.name}_success', 1)
            
            else:
                point.field(f'{event.name}_failed', 1)

            points.append(point)

        await self.write_api.write(
            bucket=self.events_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Bucket - {self.events_bucket_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Bucket - {self.shared_metrics_bucket_name}')

        points = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            point = Point(metrics_set.name)

            for tag in metrics_set.tags:
                point.tag(tag.name, tag.value)

            metric_record = {
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                **metrics_set.common_stats
            }

            for field, value in metric_record.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:common:{field}')
                point.field(field, value)

            points.append(point)

        await self.write_api.write(
            bucket=self.shared_metrics_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Bucket - {self.shared_metrics_bucket_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Bucket - {self.metrics_bucket_name}')

        points = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            
            for group_name, group in metrics_set.groups.items():
                point = Point(metrics_set.name)

                for tag in metrics_set.tags:
                    point.tag(tag.name, tag.value)

                metric_record = {
                    **group.stats, 
                    **group.custom,
                    'group': group_name
                }

                for field, value in metric_record.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:{group_name}:{field}')
                    point.field(field, value)

                points.append(point)

        await self.write_api.write(
            bucket=self.metrics_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Bucket - {self.metrics_bucket_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        points = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_metric_name, custom_metric in metrics_set.custom_metrics.items():

                point = Point(f'{metrics_set.name}_{custom_metric_name}')

                for tag in metrics_set.tags:
                    point.tag(tag.name, tag.value)
                
                metric_record = {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'custom',
                    custom_metric_name: custom_metric.metric_value
                }

                for field, value in metric_record.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metric - {metrics_set.name}:custom:{field}')
                    point.field(field, value)

                points.append(point)

        for point in points:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to Bucket - {self.custom_bucket_name}_metrics')
            await self.write_api.write(
                bucket=self.custom_bucket_name,
                record=points
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Bucket - {self.custom_bucket_name}_metrics')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Errors Metrics to Bucket - {self.errors_bucket_name}')

        points = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                point = Point(f'{metrics_set.name}_errors')
                point.field(
                    error.get('message'),
                    error.get('count')
                )

            points.append(point)

        await self.write_api.write(
            bucket=self.errors_bucket_name,
            record=points
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics to Bucket - {self.errors_bucket_name}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connectiion to InfluxDB at - {self.protocol}://{self.host} - for Organization - {self.organization}')
        
        await self.client.close()

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connectiion to InfluxDB at - {self.protocol}://{self.host} - for Organization - {self.organization}')



import uuid
from typing import Dict, List, Any
from hedra.core.engines.types.common.metadata import Metadata
from .hooks import Hooks


class Page:

    __slots__ = (
        'selector',
        'attribute',
        'x_coordinate',
        'y_coordinate',
        'frame'
    )

    def __init__(
        self, 
        selector: str=None, 
        attribute: str=None, 
        x_coordinate: int=0, 
        y_coordinate: int=0, 
        frame=0
    ) -> None:
        self.selector = selector
        self.attribute = attribute
        self.x_coordinate = x_coordinate
        self.y_coordinate = y_coordinate
        self.frame = frame


class URL:

    __slots__ = (
        'location',
        'headers'
    )

    def __init__(
        self, 
        location: str=None, 
        headers: Dict[str, str]={}
    ) -> None:
        self.location = location
        self.headers = headers


class Input:

    __slots__ = (
        'key',
        'text',
        'expression',
        'args',
        'filepath',
        'file',
        'path',
        'option',
        'by_label',
        'by_value'
    )

    def __init__(
        self, 
        key=None, 
        text=None, 
        expression: str=None, 
        args: List[Any]=None, 
        filepath=None, 
        file=None, 
        path: str=None,
        option: Any=None,
        by_label: bool=False,
        by_value: bool=False
    ) -> None:
        self.key = key
        self.text = text
        self.expression = expression
        self.args = args
        self.filepath = filepath
        self.file = file
        self.path = path
        self.option = option
        self.by_label = by_label
        self.by_value = by_value


class Options:

    __slots__ = (
        'event',
        'option',
        'is_checked',
        'timeout',
        'extra',
        'switch_by'
    )

    def __init__(
        self, 
        event: str=None, 
        option=None, 
        is_checked=False, 
        timeout=10, 
        extra: Dict[str, Any]={}, 
        switch_by: str='url'
    ) -> None:
        self.event = event
        self.option = option
        self.is_checked = is_checked
        self.timeout = timeout
        self.extra = extra
        self.switch_by = switch_by


class PlaywrightCommand:

    __slots__ = (
        'action_id'
        'name',
        'command',
        'page',
        'url',
        'input',
        'options',
        'metadata',
        'hooks',
        'event',
        'command_args',
        'mutations'
    )
    
    def __init__(self, 
        name, 
        command, 
        page: Page = Page(), 
        url: URL = URL(), 
        input: Input = Input(), 
        options: Options = Options(), 
        user: str = None, 
        tags: List[Dict[str, str]] = []
    ) -> None:
        self.action_id = str(uuid.uuid4())
        self.name = name
        self.command = command
        self.page = page
        self.url = url
        self.input = input
        self.options = options
        self.metadata = Metadata(user, tags)
        self.hooks: Hooks[PlaywrightCommand] = Hooks()
        self.event = None
        self.command_args: Dict[str, Any] = {}



import os
import click
from hedra.cli.project import (
    create_project,
    sync_project,
    get_project,
    about_project
)


@click.group(help='Commands for managing collections of Hedra graphs.')
def project():
    pass


@project.command(
    help='Creates a project at the specified path.'
)
@click.argument('url')
@click.option(
    '--project-name',
    default='hedra_tests',
    help='Name of project to create.'
)
@click.option(
    '--path',
    default=os.getcwd(),
    help='Path to graph repository.'
)
@click.option(
    '--username',
    help='Git repository username.'
)
@click.option(
    '--password',
    help='Git repository password'
)
@click.option(
    '--bypass-connection-validation',
    is_flag=True,
    show_default=True,
    default=False,
    help="Skip Hedra's action connection validation."
)
@click.option(
    '--connection-validation-retries',
    default=3,
    help="Set the number of retries for connection validation."
)
@click.option(
    '--log-level',
    default='info',
    help='Set log level.'
)
def create(
    url: str, 
    project_name: str,
    path: str,
    username: str, 
    password: str,
    bypass_connection_validation: bool,
    connection_validation_retries: int,
    log_level: str
):
    create_project(
        url,
        project_name,
        path,
        username,
        password,
        bypass_connection_validation,
        connection_validation_retries,
        log_level
    )



@project.command(
    help="Synchronize changes to the project at the specified path."
)
@click.option(
    '--url',
    help='Git repository url.'
)
@click.option(
    '--path',
    default=os.getcwd(),
    help='Path to graph repository.'
)
@click.option(
    '--branch',
    help='Git repository branch.'
)
@click.option(
    '--remote',
    help='Git repository remote.'
)
@click.option(
    '--sync-message',
    help='Message for git commit.'
)
@click.option(
    '--username',
    help='Git repository username.'
)
@click.option(
    '--password',
    help='Git repository password'
)
@click.option(
    '--ignore',
    help='Comma delimited list of files to add to the project .gitignore.'
)
@click.option(
    '--log-level',
    default='info',
    help='Set log level.'
)
@click.option(
    '--local',
    is_flag=True,
    show_default=True,
    default=False,
    help='Synchronize only local project state.'
)
def sync(
    url: str, 
    path: str,
    branch: str, 
    remote: str, 
    sync_message: str, 
    username: str, 
    password: str,
    ignore: str,
    log_level: str,
    local: bool
):
    sync_project(
        url,
        path,
        branch,
        remote,
        sync_message,
        username,
        password,
        ignore,
        log_level,
        local
    )


@project.command(
    help='Clone down remote project to the specified path'
)
@click.argument('url')
@click.option(
    '--path',
    default=os.getcwd(),
    help='Path to graph repository.'
)
@click.option(
    '--branch',
    help='Git repository branch.'
)
@click.option(
    '--remote',
    help='Git repository remote.'
)
@click.option(
    '--username',
    help='Git repository username.'
)
@click.option(
    '--password',
    help='Git repository password'
)
@click.option(
    '--bypass-connection-validation',
    is_flag=True,
    show_default=True,
    default=False,
    help="Skip Hedra's action connection validation."
)
@click.option(
    '--connection-validation-retries',
    default=3,
    help="Set the number of retries for connection validation."
)
@click.option(
    '--log-level',
    default='info',
    help='Set log level.'
)
def get(
    url: str, 
    path: str,
    branch: str, 
    remote: str, 
    username: str, 
    password: str,
    bypass_connection_validation: bool,
    connection_validation_retries: int,
    log_level: str
):
    get_project(
        url,
        path,
        branch,
        remote,
        username,
        password,
        bypass_connection_validation,
        connection_validation_retries,
        log_level
    )
 

@project.command(
    help="Describe the project at the specified path"
)
@click.option(
    '--path',
    default=os.getcwd(),
    help='Path to graph repository.'
)
def about(
    path: str
):
    about_project(path)


import uuid
import json
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.graphql import (
    GraphQLAction,
    MercuryGraphQLClient
)
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.data.parsers.parser_types.common.base_parser import BaseParser
from hedra.data.parsers.parser_types.common.parsing import (
    normalize_headers,
    parse_tags
)
from typing import Any, Coroutine, Dict
from .graphql_action_validator import GraphQLActionValidator


class GraphQLActionParser(BaseParser):

    def __init__(
        self,
        config: Config,
        options: Dict[str, Any]={}
    ) -> None:
        super().__init__(
            GraphQLActionParser.__name__,
            config,
            RequestTypes.GRAPHQL,
            options
        )

    async def parse(
        self, 
        action_data: Dict[str, Any],
        stage: str
    ) -> Coroutine[Any, Any, Coroutine[Any, Any, ActionHook]]:
        
        graphql_variables_data = action_data.get('variables')
        if isinstance(graphql_variables_data, (str, bytes, bytearray,)):
            graphql_variables_data = json.loads(graphql_variables_data)

        normalized_headers = normalize_headers(action_data)
        tags_data = parse_tags(action_data)

        generator_action = GraphQLActionValidator(**{
            **action_data,
            'headers': normalized_headers,
            'variables': graphql_variables_data,
            'tags': tags_data
        })

        action = GraphQLAction(
            generator_action.name,
            generator_action.url,
            method=generator_action.method,
            headers=generator_action.headers,
            data={
                'query': generator_action.query,
                'operation_name': generator_action.operation_name,
                'variables': generator_action.variables
            },
            user=generator_action.user,
            tags=[
                tag.dict() for tag in generator_action.tags
            ]
        )

        session = MercuryGraphQLClient(
            concurrency=self.config.batch_size,
            timeouts=self.timeouts,
            reset_connections=self.config.reset_connections,
            tracing_session=self.config.tracing
        )

        await session.prepare(action)

        hook = ActionHook(
            f'{stage}.{generator_action.name}',
            generator_action.name,
            None,
            sourcefile=generator_action.sourcefile
        )

        hook.session = session
        hook.action = action
        hook.stage = stage
        hook.context = SimpleContext()
        hook.hook_id = uuid.uuid4()
      
        hook.metadata.order = generator_action.order
        hook.metadata.weight = generator_action.weight
        hook.metadata.tags = generator_action.tags
        hook.metadata.user = generator_action.user

        return hook






from typing import Optional
from hedra.reporting.types.common.types import ReporterTypes
from pydantic import BaseModel


class RedisConfig(BaseModel):
    host: str='localhost:6379'
    username: Optional[str]
    password: Optional[str]
    database: int=0
    events_channel: str='events'
    metrics_channel: str='metrics'
    experiments_channel: str='experiments'
    streams_channel: str='streams'
    system_metrics_channel: str='system_metrics'
    channel_type: str='pipeline'
    secure: bool=False
    reporter_type: ReporterTypes=ReporterTypes.Redis


import uuid
from typing import Any, List, Dict, Union
from hedra.core.hooks.types.base.event import BaseEvent
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.reporting.tags import Tag


class BaseProcessedResult:

    __slots__ = (
        'stage_name',
        'event_id',
        'action_id',
        'name',
        'shortname',
        'error',
        'type',
        'source',
        'checks',
        'tags',
        'stage',
        'timings',
        'time'
    )

    def __init__(
        self, 
        stage_name: str, 
        result: BaseResult
    ) -> None:

        self.event_id = str(uuid.uuid4())
        self.action_id = result.action_id

        self.name = result.name
        self.shortname = result.name
        self.error = result.error
        self.timings = {}
        self.type = result.type
        self.source = result.source
        self.checks: List[List[BaseEvent]] = []
        self.stage = stage_name

        self.time = self.timings.get('total', 0)

        self.tags = [
            Tag(
                tag.get('name'),
                tag.get('value')
            ) for tag in result.tags
        ]


    @property
    def fields(self):
        return ['id', 'name', 'stage', 'time', 'succeeded']

    @property
    def values(self):
        return [self.name, self.stage, self.time, self.success]

    @property
    def record(self):
        return {
            'id': self.event_id,
            'name': self.name,
            'stage': self.stage,
            'time': self.time,
            'succeeded': self.success
        }

    @property
    def success(self):
        return self.error is None

    def tags_to_dict(self):
        return {
            tag.name: tag.value for tag in self.tags
        }
    
    def to_dict(self) -> Dict[str, Union[str, int, float]]:
        raise NotImplementedError('Err. - Implement this method in a specific class inheriting from BaseProcessedResult')

    def serialize(self) -> str:
        raise NotImplementedError('Err. - Implement this method in a specific class inheriting from BaseProcessedResult')


import math
import asyncio
import psutil
from hedra.core.personas.types.default_persona.default_persona import DefaultPersona
from hedra.core.personas.batching.param_type import ParamType
from hedra.core.engines.client.config import Config
from hedra.core.personas.batching.batch import Batch
from hedra.core.graphs.stages.optimize.optimization.parameters.parameter import Parameter
from typing import (
    Dict, 
    List, 
    Tuple, 
    Union, 
    Optional
)


class BaseAlgorithm:

    def __init__(
        self, 
        config: Dict[str, Union[List[Tuple[Union[int, float]]], int, Config]],
        distribution_idx: int=None
    ) -> None:
        self.stage_config: Config = config.get('stage_config')
        self.max_iter = config.get('iterations', 10)
        self.distribution: Optional[List[int]] = None

        if self.stage_config.experiment:
            self.distribution = self.stage_config.experiment.get('distribution')

        parameters: List[Parameter] = config.get('params', [])
        algorithm_parameters: Dict[str, Parameter] = {}

        for parameter in parameters:
            algorithm_parameters[parameter.parameter_name] = parameter

        self.params: Dict[str, Parameter] = algorithm_parameters
        self.time_limit = config.get('time_limit', 60)
        self.persona_total_time = self.stage_config.total_time
        self.batch = Batch(self.stage_config)
        self.current_params = {}
        self.session = None
        self.distribution_idx: Optional[int] = distribution_idx
        self._boundaries_by_param_name = {}

        if self.max_iter is None or self.max_iter <= 0:
            self.max_iter = 10

        if self.time_limit is None or self.time_limit <= 0:
            self.time_limit = 60

        self.batch_time = self.time_limit/self.max_iter

        self.param_names = []
        self.bounds = []

        self.param_values: Dict[str, Union[float, int]] = self.get_params()
        self.current_params = {}

        self.optimize_params = [
            param_name for param_name in self.params.keys() if param_name in self.param_values
        ]

        if len(self.optimize_params) < 1:
            self.optimize_params = [
                'batch_size'
            ]

        for param_name in self.optimize_params:
            parameter = self.params.get(param_name)
                    
            param = self.param_values.get(param_name, {})
            value = param.get('value')
            params_type = param.get('type')

            if self.distribution_idx is not None:
                parameter.minimum = 0.01
                parameter.maximum = 10
                value = self.distribution[self.distribution_idx]
                param['value'] = value

            min_range = parameter.minimum
            max_range = parameter.maximum

            self.param_names.append(param_name)

            if parameter.feed_forward:
                if params_type == ParamType.INTEGER:
                    min_range = math.floor(min_range * value)
                    max_range = math.ceil(max_range * value)

                else:
                    min_range = min_range * value
                    max_range = max_range * value

            boundaries = (min_range, max_range)
            self.bounds.append(boundaries)
            self._boundaries_by_param_name[param_name] = boundaries

            self.current_params[param_name] = value
            
        self.fixed_iters = False
        self.iters = 0

    def get_params(self):
        return self.batch.to_params()

    def update_params(self, persona: DefaultPersona) -> DefaultPersona:

        persona.total_time = self.batch_time

        batch_size = int(self.current_params.get(
            'batch_size',
            persona.batch.size
        ))

        batch_interval = float(self.current_params.get(
            'batch_interval',
            persona.batch.interval
        ))

        batch_gradient = float(self.current_params.get(
            'batch_gradient',
            persona.batch.gradient
        ))

        for hook in persona._hooks:

            if batch_size <= psutil.cpu_count():
                batch_size = 1000

            hook.session.pool.size = batch_size
            hook.session.sem = asyncio.Semaphore(batch_size)
            hook.session.pool.connections = []
            hook.session.pool.create_pool()

        persona.batch.size = batch_size
        persona.batch.interval = batch_interval
        persona.batch.gradient = batch_gradient

        return persona

    def optimize(self, func):
        raise NotImplementedError(
            'Err. - Base Algorithm is an abstract class and its optimize method should be overridden.'
        )


from typing import List, Dict
from .stream import Stream


class StreamAnalytics:

    __slots__ = (
        'interval_completion_rates',
        'interval_timings',
        'interval_completed_counts',
        'interval_succeeded_counts',
        'interval_failed_counts',
        'interval_batch_timings'
    )

    def __init__(self) -> None:    
        self.interval_completion_rates: List[float] = []
        self.interval_timings: List[Dict[str, float]] = []
        self.interval_completed_counts: List[int] = []
        self.interval_succeeded_counts: List[int] = []
        self.interval_failed_counts: List[int] = []
        self.interval_batch_timings: List[float] = []

    def add(
            self, 
            stream: Stream,
            batch_elapsed: float
        ):

        completed_count = len(stream.completed)
        self.interval_completion_rates.append(
            completed_count/batch_elapsed
        )

        self.interval_completed_counts.append(completed_count)
        self.interval_succeeded_counts.append(stream.succeeded)
        self.interval_failed_counts.append(stream.failed)
        self.interval_timings.append(stream.timings)
        self.interval_batch_timings.append(batch_elapsed)

    




from __future__ import annotations
from hedra.core.engines.types.tracing.tracing_types import (
    RequestHook,
    ResponseHook,
    TraceSignal,
    UrlFilter
)
from hedra.core.engines.types.tracing.url_filters import default_params_strip_filter
from hedra.versioning.flags.types.unstable.flag import unstable
from typing import Optional, Dict, Union


OpenTelemetryTracingConfig = Union[
    UrlFilter,
    RequestHook,
    ResponseHook,
    TraceSignal
]


class TracingConfig:

    def __init__(
        self,
        url_filter: Optional[UrlFilter]=None,
        request_hook: Optional[RequestHook]=None,
        response_hook: Optional[ResponseHook]=None,
        on_request_headers_sent: Optional[TraceSignal]=None,
        on_request_data_sent: Optional[TraceSignal]=None,
        on_request_chunk_sent: Optional[TraceSignal]=None,
        on_response_headers_received: Optional[TraceSignal]=None,
        on_response_data_received: Optional[TraceSignal]=None,
        on_response_chunk_received: Optional[TraceSignal]=None,
        on_request_redirect: Optional[TraceSignal]=None,
        on_connection_queued_start: Optional[TraceSignal]=None,
        on_connection_queued_end: Optional[TraceSignal]=None,
        on_connection_create_start: Optional[TraceSignal]=None,
        on_connection_create_end: Optional[TraceSignal]=None,
        on_connection_reuse_connection: Optional[TraceSignal]=None,
        on_dns_resolve_host_start: Optional[TraceSignal]=None,
        on_dns_resolve_host_end: Optional[TraceSignal]=None,
        on_dns_cache_hit: Optional[TraceSignal]=None,
        on_dns_cache_miss: Optional[TraceSignal]=None,
        on_task_start: Optional[TraceSignal]=None,
        on_task_end: Optional[TraceSignal]=None,
        on_task_error: Optional[TraceSignal]=None
    ) -> None:

        if url_filter is None:
            url_filter = default_params_strip_filter

        self.url_filter = url_filter
        self.request_hook = request_hook
        self.response_hook = response_hook

        self.on_request_headers_sent: TraceSignal = on_request_headers_sent
        self.on_request_data_sent: TraceSignal = on_request_data_sent
        self.on_request_chunk_sent: TraceSignal = on_request_chunk_sent
        self.on_request_redirect: TraceSignal = on_request_redirect
        self.on_response_headers_received = on_response_headers_received
        self.on_response_data_received = on_response_data_received
        self.on_response_chunk_received: TraceSignal = on_response_chunk_received

        self.on_connection_queued_start: TraceSignal = on_connection_queued_start
        self.on_connection_queued_end: TraceSignal = on_connection_queued_end
        self.on_connection_create_start: TraceSignal = on_connection_create_start
        self.on_connection_create_end: TraceSignal = on_connection_create_end
        self.on_connection_reuse_connection: TraceSignal = on_connection_reuse_connection

        self.on_dns_resolve_host_start: TraceSignal = on_dns_resolve_host_start
        self.on_dns_resolve_host_end: TraceSignal = on_dns_resolve_host_end
        self.on_dns_cache_hit: TraceSignal = on_dns_cache_hit
        self.on_dns_cache_miss: TraceSignal = on_dns_cache_miss

        self.on_task_start = on_task_start
        self.on_task_end = on_task_end
        self.on_task_error = on_task_error

    def copy(self) -> TracingConfig:
        return TracingConfig(
            url_filter=self.url_filter,
            request_hook=self.request_hook,
            response_hook=self.response_hook,
            on_request_headers_sent=self.on_request_chunk_sent,
            on_request_data_sent=self.on_request_data_sent,
            on_request_chunk_sent=self.on_request_chunk_sent,
            on_request_redirect=self.on_request_redirect,
            on_response_headers_received=self.on_response_headers_received,
            on_response_data_received=self.on_response_data_received,
            on_response_chunk_received=self.on_response_chunk_received,
            on_connection_queued_start=self.on_connection_queued_start,
            on_connection_queued_end=self.on_connection_queued_end,
            on_connection_create_start=self.on_connection_create_start,
            on_connection_create_end=self.on_connection_create_end,
            on_connection_reuse_connection=self.on_connection_reuse_connection,
            on_dns_resolve_host_start=self.on_dns_resolve_host_start,
            on_dns_resolve_host_end=self.on_dns_resolve_host_end,
            on_dns_cache_hit=self.on_dns_cache_hit,
            on_dns_cache_miss=self.on_dns_cache_miss,
            on_task_start=self.on_task_start,
            on_task_end=self.on_task_end,
            on_task_error=self.on_task_error,
        )
    
    def to_dict(self) -> Dict[str, OpenTelemetryTracingConfig]:
        return {
            'url_filter': self.url_filter,
            'request_hook': self.request_hook,
            'response_hook': self.response_hook,
            'on_request_headers_sent': self.on_request_chunk_sent,
            'on_request_data_sent': self.on_request_data_sent,
            'on_request_chunk_sent': self.on_request_chunk_sent,
            'on_request_redirect': self.on_request_redirect,
            'on_response_headers_received': self.on_response_headers_received,
            'on_response_data_received': self.on_response_data_received,
            'on_response_chunk_received': self.on_response_chunk_received,
            'on_connection_queued_start': self.on_connection_queued_start,
            'on_connection_queued_end': self.on_connection_queued_end,
            'on_connection_create_start': self.on_connection_create_start,
            'on_connection_create_end': self.on_connection_create_end,
            'on_connection_reuse_connection': self.on_connection_reuse_connection,
            'on_dns_resolve_host_start': self.on_dns_resolve_host_start,
            'on_dns_resolve_host_end': self.on_dns_resolve_host_end,
            'on_dns_cache_hit': self.on_dns_cache_hit,
            'on_dns_cache_miss': self.on_dns_cache_miss,
            'on_task_start': self.on_task_start,
            'on_task_end': self.on_task_end,
            'on_task_error': self.on_task_error,
        }



import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    InfluxDBConfig
)


class SubmitInfluxDBResultsStage(Submit):
    config=InfluxDBConfig(
        host='localhost:8006',
        token=os.getenv('INFLUXDB_API_TOKEN', ''),
        organization='<organization_here>',
        events_bucket='events',
        metrics_bucket='metrics'
    )


import statistics
import uuid
from typing import Union, List, Dict, Any
from hedra.core.experiments.distribution_types import DistributionTypes
from hedra.reporting.metric import MetricsSet
from .experiments_collection import ExperimentMetricsCollection
from .experiment_metrics_set_types import (
    MutationSummary,
    VariantSummary,
    ExperimentSummary
)


RawSummaryItem = Dict[str, Union[str, int, float, bool, List[float]]]


class ExperimentMetricsSet:

    experiments_table_header_keys = [
        'experiment_name',
        'experiment_randomized',
        'experiment_completed',
        'experiment_succeeded',
        'experiment_failed',
        'experiment_median_aps'
    ]

    variants_table_header_keys = [
        'variant_name',
        'variant_experiment',
        'variant_weight',
        'variant_distribution',
        'variant_distribution_interval',
        'variant_ratio_completed',
        'variant_ratio_succeeded',
        'variant_ratio_failed',
        'variant_ratio_aps',
    ]

    variants_stats_table_header_keys = [
        'variant_name',
        'variant_completed',
        'variant_succeeded',
        'variant_failed',
        'variant_actions_per_second',
    ]
    
    mutations_table_headers_keys = [
        'mutation_name',
        'mutation_experiment_name',
        'mutation_variant_name',
        'mutation_chance',
        'mutation_targets',
        'mutation_type'
    ]

    def __init__(self) -> None:
        self.experiment_metrics_set_id = uuid.uuid4()
        self.experiment_name: Union[str, None] = None
        self.randomized: Union[bool, None] = None
        self.participants: List[str] = []
        self.variants: Dict[str, Dict[str, Any]] = {}
        self.mutations: Dict[str, List[Dict[str, Any]]] = {}
        self.metrics: Dict[str, Dict[str, MetricsSet]] = {}
        
        self.experiments_table_headers: Dict[str, str] = {}
        self.variants_table_headers: Dict[str, str] = {}
        self.mutations_table_headers: Dict[str, str] = {}
        self.mutations_summaries: Dict[str, str] = {}

        self.experiments_summary: ExperimentSummary = {}

    @classmethod
    def experiments_fields(cls):
        return cls.experiments_table_header_keys
    
    @classmethod
    def variants_fields(cls):
        return list(set([
            *cls.variants_table_header_keys,
            *cls.variants_stats_table_header_keys
        ]))
    
    @classmethod
    def mutations_fields(cls):
        return cls.mutations_table_headers_keys

    def generate_variant_summaries(self) -> Dict[str, VariantSummary]:

        variants_summaries = {}

        for participant in self.participants:
            variant = self.variants.get(participant)
            mutations = self.mutations.get(participant)
            metrics_set = self.metrics.get(participant)

            group_variant_summaries: Dict[str, Union[int, List[float]]] = {
                'variant_completed': 0,
                'variant_succeeded': 0,
                'variant_failed': 0,
                'variant_actions_per_second': []
            }
            for variant_metric_set in metrics_set.values():
                group_variant_summaries['variant_completed'] += variant_metric_set.common_stats.get('total')
                group_variant_summaries['variant_succeeded'] += variant_metric_set.common_stats.get('succeeded')
                group_variant_summaries['variant_failed'] += variant_metric_set.common_stats.get('failed')
                group_variant_summaries['variant_actions_per_second'].append(
                    variant_metric_set.common_stats.get('actions_per_second')
                )

            group_variant_summaries['variant_actions_per_second'] = round(
                statistics.median(
                    group_variant_summaries.get('variant_actions_per_second')
                ), 2
            )

            variant_distribution_type: DistributionTypes = variant.get('variant_distribution_type')

            variant_summary = VariantSummary(**{
                'variant_name': participant,
                'variant_weight': variant.get('variant_weight'),
                'variant_experiment': '',
                'variant_distribution': variant_distribution_type.name.capitalize(),
                'variant_distribution_interval': variant.get('variant_distribution_interval_duration'),
                'variant_mutation_summaries': {},
                **group_variant_summaries,
            })


            if len(mutations) > 0:

                for mutation in mutations:

                    mutation_name = mutation.get('mutation_name')

                    mutations_summary = MutationSummary(**{
                        'mutation_experiment_name': self.experiment_name,
                        'mutation_variant_name': participant,
                        'mutation_name': mutation_name,
                        'mutation_chance': mutation.get('mutation_chance'),
                        'mutation_targets': ':'.join(mutation.get('mutation_targets')),
                        'mutation_type': mutation.get('mutation_type')
                    })

                    self.mutations_table_headers.update({
                        header_name: header_name.replace(
                            'mutation_', ''
                        ).replace(
                            '_', ' '
                        ) for header_name in mutations_summary.dict().keys()
                    })
                        
                    self.mutations_summaries[mutation_name] = mutations_summary

                    variant_summary.variant_mutation_summaries[mutation_name] = mutations_summary

            variants_summaries[participant] = variant_summary

        return variants_summaries
    
    def generate_experiment_summary(self) -> None:
        variant_summaries = self.generate_variant_summaries()

        experiment_summary = ExperimentSummary(**{
            'experiment_name': self.experiment_name,
            'experiment_randomized': self.randomized,
            'experiment_completed': 0,
            'experiment_succeeded': 0,
            'experiment_failed': 0,
            'experiment_median_aps': 0.,
            'experiment_variant_summaries': {}
        })

        experiment_actions_per_second = []

        for participant in self.participants:

            selected_variant_summary = variant_summaries.get(participant)
            selected_variant_summary.variant_experiment = self.experiment_name

            variant_stats: Dict[str, List[Union[int, float]]] = {
                'alt_variants_completed': [],
                'alt_variants_succeeded': [],
                'alt_variants_failed': [],
                'alt_variants_actions_per_second': [],
            }

            for variant_name, variant_summary in variant_summaries.items():

                if variant_name != participant:

                    variant_stats['alt_variants_completed'].append(
                        variant_summary.variant_completed
                    )

                    variant_stats['alt_variants_succeeded'].append(
                        variant_summary.variant_succeeded
                    )

                    variant_stats['alt_variants_failed'].append(
                        variant_summary.variant_failed
                    )

                    variant_stats['alt_variants_actions_per_second'].append(
                        variant_summary.variant_actions_per_second
                    )
            
            variant_total_completed = selected_variant_summary.variant_completed
            mean_alt_variants_total_completed = statistics.mean(
                variant_stats.get('alt_variants_completed')
            )

            selected_variant_summary.variant_ratio_completed= round(
                variant_total_completed/mean_alt_variants_total_completed,
                2
            )
            
            variant_total_succeeded = selected_variant_summary.variant_succeeded
            mean_alt_variants_total_succeeded = statistics.mean(
                variant_stats.get('alt_variants_succeeded')
            )

            selected_variant_summary.variant_ratio_succeeded = round(
                variant_total_succeeded/mean_alt_variants_total_succeeded,
                2
            )

            variant_total_failed = selected_variant_summary.variant_failed
            mean_alt_variants_total_failed = statistics.mean(
                variant_stats.get('alt_variants_failed')
            )

            selected_variant_summary.variant_ratio_failed = round(
                variant_total_failed/mean_alt_variants_total_failed,
                2
            )

            variant_total_actions_per_second = selected_variant_summary.variant_actions_per_second
            median_alt_variants_actions_per_second = statistics.median(
                variant_stats.get('alt_variants_actions_per_second')
            )

            selected_variant_summary.variant_ratio_aps = round(
                variant_total_actions_per_second/median_alt_variants_actions_per_second,
                2
            )

            self.variants_table_headers.update({
                header_name: header_name.replace(
                    'variant_', ''
                ).replace(
                    '_', ' '
                ) for header_name in selected_variant_summary.dict().keys() if 'mutation' not in header_name
            })

            experiment_summary.experiment_variant_summaries[participant] = selected_variant_summary

            experiment_summary.experiment_completed += selected_variant_summary.variant_completed
            experiment_summary.experiment_succeeded += selected_variant_summary.variant_succeeded
            experiment_summary.experiment_failed += selected_variant_summary.variant_failed
            
            experiment_actions_per_second.append(
                selected_variant_summary.variant_actions_per_second
            )

        experiment_summary.experiment_median_aps = statistics.median(experiment_actions_per_second)

        self.experiments_table_headers = {
            header_name: header_name.replace(
                'experiment_', ''
            ).replace(
                '_', ' '
            ) for header_name in experiment_summary.dict().keys() if 'variant' not in header_name
        }
        
        self.experiments_summary = experiment_summary

    
    def split_experiments_metrics(self) -> ExperimentMetricsCollection:

        variant_records: List[Dict[str, str]] = []
        mutations_records: List[Dict[str, str]] = []
        mutations_summaries: List[MutationSummary] = []

        for variant in self.experiments_summary.experiment_variant_summaries.values():
            variant_records.append(
                variant.dict(
                    include={header for header in self.variants_fields()}
                )
            )

            for mutation in variant.variant_mutation_summaries.values():
                mutations_records.append(
                    mutation.dict(
                        include={header for header in self.mutations_table_headers_keys}
                    )
                )

                mutations_summaries.append(mutation)

        return ExperimentMetricsCollection(
            experiment=self.experiments_summary.dict(
                include={header for header in self.experiments_table_header_keys}
            ),
            variants=variant_records,
            mutations=mutations_records,
            experiment_summary=self.experiments_summary,
            variant_summaries=list(self.experiments_summary.experiment_variant_summaries.values()),
            mutation_summaries=mutations_summaries
        )




from hedra.distributed.discovery.dns.core.cache import CacheNode
from hedra.distributed.models.dns import (
    DNSMessage,
    QueryType
)
from hedra.distributed.discovery.dns.core.config import core_config
from hedra.distributed.discovery.dns.core.nameservers import NameServer
from hedra.distributed.discovery.dns.core.record import (
    Record,
    RecordType,
    RecordTypesMap
)
from hedra.distributed.env import Env
from typing import Callable, List, Tuple, Optional, Union
from .base_resolver import BaseResolver
from .memoizer import Memoizer



Proxy = Tuple[
    Union[
        Callable[
            [str],
            bool
        ],
        str,
        None
    ], 
    str
]

NameServerPair = Tuple[
    Union[
        Callable[
            [str],
            bool
        ],
        None
    ],
    NameServer
]


class ProxyResolver(BaseResolver):

    default_nameservers = core_config['default_nameservers']
    memoizer = Memoizer()

    def __init__(
        self, 
        host: str,
        port: int,
        instance_id: str,
        env: Env,
        cache: CacheNode = None,
        proxies: Optional[List[Proxy]]=None
    ):
        super().__init__(
            host,
            port,
            instance_id,
            env,
            cache=cache
        )

        if proxies is None:
            proxies = self.default_nameservers

        self.types_map = RecordTypesMap()
        self._nameserver_pairs = self.set_proxies(proxies)

    def _get_matching_nameserver(self, fqdn):

        for nameserver_test, nameserver in self._nameserver_pairs:
            if nameserver_test is None or nameserver_test(fqdn):
                return nameserver

        return NameServer([])
    
    def add_nameserver(
        self,
        urls: List[str]
    ):
        namserver = NameServer(urls)

        self._nameserver_pairs.append((
            None,
            namserver
        ))

        return namserver.data

    @staticmethod
    def build_tester(rule) ->  Callable[
        [str],
        bool
    ]:

        if rule is None or callable(rule):
            return rule
        
        assert isinstance(rule, str)

        if rule.startswith('*.'):
            suffix = rule[1:]

            return lambda d: d.endswith(suffix)
        
        return lambda d: d == rule

    def set_proxies(
        self, 
        proxies: List[Proxy]
    ):

        nameserver_pairs: List[NameServerPair] = []
        fallback: List[str] = []

        if proxies:
            for item in proxies:

                if isinstance(item, str):
                    fallback.append(item)
                    continue

                test, nameserver = item
                if test is None:
                    fallback.extend(nameserver)
                    continue

                nameserver_pairs.append(
                    (
                        self.build_tester(test), 
                        NameServer([nameserver])
                    )
                )

        if fallback:
            nameserver_pairs.append(
                (
                    None, 
                    NameServer(fallback)
                )
            )

        return nameserver_pairs

    @memoizer.memoize_async(
            lambda _, fqdn, record_type, skip_cache: (fqdn, record_type)
    )
    async def _query(
        self, 
        fqdn: str, 
        record_type: RecordType,
        skip_cache: bool
    ):

        msg = DNSMessage()
        msg.query_domains.append(
            Record(
                QueryType.REQUEST, 
                name=fqdn, 
                record_type=record_type
            )
        )

        has_result = False

        if skip_cache is False:
            has_result, fqdn = self.query_cache(msg, fqdn, record_type)

        while not has_result:
            nameserver = self._get_matching_nameserver(fqdn)

            for addr in nameserver.iter():
                try:
                    res = await self.request(fqdn, msg, addr)

                except:
                    nameserver.fail(addr)
                    raise

                else:
                    nameserver.success(addr)
                    self.cache_message(res)
                    msg.query_answers.extend(res.query_answers)
                    has_result = True

                    break

        return msg




class UnsafeNotEnabledException(Exception):

    def __init__(self, feature_name: str) -> None:
        super().__init__(
            f'\nErr. - Attempting to use unsafe feature - {feature_name} - wihtout --enable-unsafe flag.\nPlease pass this flag if you want to use unstable features.\n'
        )


from enum import Enum
from hedra.distributed.models.http import (
    Request,
    Response
)
from pydantic import BaseModel
from typing import (
    Any,
    Callable, 
    Union, 
    Tuple,
    Coroutine
)


class MiddlewareType(Enum):
    BIDIRECTIONAL='BIDIRECTIONAL'
    CALL='CALL'
    UNIDIRECTIONAL_BEFORE='UNIDIRECTIONAL_BEFORE'
    UNIDIRECTIONAL_AFTER='UNIDIRECTIONAL_AFTER'


RequestHandler = Callable[
    [Request],
    Coroutine[
        Any,
        Any,
        Tuple[
            Union[
                Response,
                BaseModel,
                str,
                None
            ],
            int
        ]
    ]
]

WrappedHandler = Callable[
    [
        Request,
        Response,
        int
    ],
    Coroutine[
        Any,
        Any,
        Tuple[
            Response,
            int
        ]
    ]
]

CallHandler = Callable[
    [
        Request,
        RequestHandler
    ],
    Coroutine[
        Any, 
        Any, 
        Tuple[
            Request,
            Response, 
            int
        ]
    ]
]

MiddlewareHandler = Callable[
    [
        Request,
        Response,
        int
    ],
    Coroutine[
        Any, 
        Any, 
        Tuple[
            Tuple[Response, int],
            bool
        ]
    ]
]



BidirectionalMiddlewareHandler = Callable[
    [
        Request,
        Response,
        int
    ],
    Coroutine[
        Any, 
        Any, 
        Tuple[
            Tuple[
                Request,
                Response, 
                int
            ],
            bool
        ]
    ]
]



Handler = Union[RequestHandler, WrappedHandler]


from typing import Dict, Union, List
from collections import defaultdict
from hedra.reporting.experiment.experiment_metrics_set import ExperimentSummary
from hedra.reporting.system.system_metrics_set import SystemMetricsSet
from .execution_summary_table import ExecutionSummaryTable
from .experiments_summary_table import ExperimentsSummaryTable
from .system_summary_table import SystemSummaryTable
from .table_types import (
    GraphExecutionResults,
    ExecutionResults,
    SystemMetricsCollection,
    GraphResults
)






class SummaryTable:

    def __init__(
        self,
        graph_results: GraphResults,
        summaries_visibility_config: Dict[str, bool]={}
    ) -> None:
        self.experiment_summary_table: Union[ExperimentsSummaryTable, None] = None
        self.execution_summary_table: Union[ExecutionSummaryTable, None] = None
        self.system_summary_table = SystemSummaryTable()
        self.system_summary_table.enabled_tables.update({
            table_name: enabled for table_name, enabled in summaries_visibility_config.items() if table_name in self.system_summary_table.enabled_tables
        })   

        graph_execution_results: GraphExecutionResults = graph_results.get('metrics')
        submit_stage_system_metrics: SystemMetricsCollection = graph_results.get('submit_stage_system_metrics', {})
        graph_system_metrics: SystemMetricsSet = graph_results.get('graph_system_metrics')

        self.system_summary_table.graph_metrics_summary = graph_system_metrics

        for system_metrics in submit_stage_system_metrics.values():
            self.system_summary_table.system_metrics_summaries.append(system_metrics)

        experiment_summaries: Dict[str, ExperimentSummary] = {}
        execution_results: ExecutionResults = {}

        experiment_headers: Dict[str, Dict[str]] = defaultdict(dict)
        experiment_headers_keys: Dict[str, List[str]] = defaultdict(list)

        for results_set in graph_execution_results.values():
            stage_execution_results: ExecutionResults = results_set.get('stages')
            execution_results.update(stage_execution_results)

            system_metrics: SystemMetricsSet = results_set.get('system_metrics')
            self.system_summary_table.system_metrics_summaries.append(system_metrics)
            
            experiment_metrics_sets = results_set.get('experiment_metrics_sets', {})

            for experiment_name, experiment in experiment_metrics_sets.items():
                experiment_summaries[experiment_name] = experiment.experiments_summary

                experiment_headers['experiment_table_headers'].update(
                    experiment.experiments_table_headers
                )

                experiment_headers['variants_table_headers'].update(
                    experiment.variants_table_headers
                )

                experiment_headers['mutations_table_headers'].update(
                    experiment.mutations_table_headers
                )

                if len(experiment_headers_keys['experiments_table_headers_keys']) == 0:
                    experiment_headers_keys['experiments_table_headers_keys'].extend(
                        experiment.experiments_table_header_keys
                    )

                if len(experiment_headers_keys['variant_table_headers_keys']) == 0:
                    experiment_headers_keys['variant_table_headers_keys'].extend(
                        experiment.variants_table_header_keys
                    )

                if len(experiment_headers_keys['variants_stats_table_header_keys']) == 0:
                    experiment_headers_keys['variants_stats_table_header_keys'].extend(
                        experiment.variants_stats_table_header_keys
                    )

                if len(experiment_headers_keys['mutations_table_header_keys']) == 0:
                    experiment_headers_keys['mutations_table_header_keys'].extend(
                        experiment.mutations_table_headers_keys
                    )

        self.execution_summary_table = ExecutionSummaryTable(execution_results)
        self.execution_summary_table.enabled_tables.update({
            table_name: enabled for table_name, enabled in summaries_visibility_config.items() if table_name in self.execution_summary_table.enabled_tables
        })   

        if len(experiment_summaries) > 0:
            self.experiment_summary_table = ExperimentsSummaryTable(
                experiment_summaries,
                experiment_headers,
                experiment_headers_keys
            )
            
            self.experiment_summary_table.enabled_tables.update({
                table_name: enabled for table_name, enabled in summaries_visibility_config.items() if table_name in self.experiment_summary_table.enabled_tables
            })
        
    def generate_tables(self):
        self.execution_summary_table.generate_tables()
        self.system_summary_table.generate_tables()

        if self.experiment_summary_table:
            self.experiment_summary_table.generate_tables()

    def show_tables(self):

        self.execution_summary_table.show_tables()

        if self.experiment_summary_table:
            self.experiment_summary_table.show_tables()

        self.system_summary_table.show_tables()


import json
import string
import random
from hedra.core.engines.types.graphql.action import GraphQLAction
from hedra.core.engines.types.graphql_http2.action import GraphQLHTTP2Action
from hedra.core.engines.types.grpc.action import GRPCAction
from hedra.core.engines.types.http.action import HTTPAction
from hedra.core.engines.types.http2.action import HTTP2Action
from hedra.core.engines.types.http3.action import HTTP3Action
from hedra.core.experiments.mutations.types.base.mutation import Mutation
from hedra.core.experiments.mutations.types.base.mutation_type import MutationType
from typing import Union, Optional, Dict, Tuple
from .validator import SmuggleRequestValidator


Request = Union[
    GraphQLAction,
    GraphQLHTTP2Action,
    GRPCAction,
    HTTPAction,
    HTTP2Action,
    HTTP3Action
]


class SmuggleRequest(Mutation):

    def __init__(
        self, 
        name: str, 
        chance: float,
        *targets: Tuple[str, ...],
        request_size: Optional[int]=None,
        smuggled_request: Optional[Union[Dict[str, str], bytes, str]]=None

    ) -> None:
        super().__init__(
            name, 
            chance,
            MutationType.SMUGGLE_REQUEST,
            *targets
        )

        validated_mutation = SmuggleRequestValidator(
            request_size=request_size,
            smuggled_request=smuggled_request
        )

        self.request_size = validated_mutation.request_size
        self.smuggled_request = validated_mutation.smuggled_request

        encoded_smuggled_request = self.smuggled_request

        if isinstance(encoded_smuggled_request, dict):
            encoded_smuggled_request = json.dumps(encoded_smuggled_request)
        
        if isinstance(encoded_smuggled_request, str):
            encoded_smuggled_request = encoded_smuggled_request.encode()

        self._encoded_smuggled_request = encoded_smuggled_request
        self._character_pool = ''.join([
            string.ascii_letters,
            string.digits,
            string.hexdigits,
            string.octdigits,
            string.punctuation,
            string.whitespace
        ])

        self.original_data: bytes = None

    async def mutate(self, action: Request=None) -> Request:

        chance_roll = random.uniform(0, 1)
        if chance_roll <= self.chance:
            return action

        if action.method not in ['POST', 'PUT', 'PATCH']:
            return action
        
        lowered_headers = {
            header_name.lower(): header_value for header_name, header_value in action.headers.items()
        }

        update_headers = {}
        if lowered_headers.get('content-type') is None:
            update_headers['Content-Type'] = len(action.encoded_data)

        if lowered_headers.get('transfer-encoding') is None:
            update_headers['Transfer-Encoding'] = 'chunked'
            
        action._headers.update(update_headers)
        action._header_items = list(action._headers.items())

        action._setup_headers()

        if self.smuggled_request is None:
            encoded_smuggled_request: str = ''.join(
                random.choices(
                    self._character_pool, 
                    k=self.request_size
                )
            )

            self._encoded_smuggled_request = encoded_smuggled_request.encode()

        if self.original_data is None:
            self.original_data = action.encoded_data
        
        action.encoded_data = self.original_data + self._encoded_smuggled_request  

        return action
    
    def copy(self):
        return SmuggleRequest(
            self.name,
            self.chance,
            *list(self.targets),
            request_size=self.request_size,
            smuggled_request=self.smuggled_request
        )


import functools
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.registrar import registrar
from .validator import TransformHookValidator


@registrar(HookType.TRANSFORM)
def transform(
    *names, 
    order: int=1,
    skip: bool=False
):
    
    TransformHookValidator(
        names=names,
        order=order,
        skip=skip
    )

    def wrapper(func):

        @functools.wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return decorator

    return wrapper


from collections import (
    OrderedDict,
    defaultdict
)
from tabulate import tabulate
from typing import Dict, List, Union
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiment_metrics_set_types import (
    ExperimentSummary
)



class ExperimentsSummaryTable:

    def __init__(
        self, 
        experiments_summaries: Dict[str, ExperimentSummary],
        experiment_headers: Dict[str, Dict[str, str]],
        experiment_headers_keys: Dict[str, List[str]]

    ) -> None:
        self.experiments_summaries = experiments_summaries
        self.experiment_headers = experiment_headers
        self.experiment_header_keys = experiment_headers_keys

        self.experiments_table: Union[str, None] = None
        self.variants_tables: Dict[str, Union[str, None]] = None
        self.variants_table_rows: Dict[str, List[OrderedDict]] = defaultdict(list)
        self.mutations_table: Union[str, None] = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.enabled_tables = {
            'experiments': False,
            'variants': False,
            'mutations': False
        }


    def generate_tables(self):
        self.experiments_table = self._generate_experiments_table()
        self.variants_tables = self._generate_variants_tables()
        self.mutations_table = self._generate_mutations_table()

    def show_tables(self):
        
        if any(self.enabled_tables.values()):
            self.logger.console.sync.info('\n-- Experiments --')

        if self.enabled_tables.get('experiments'):
            self.logger.console.sync.info('\nExperiments:\n')
            self.logger.console.sync.info(f'''{self.experiments_table}\n''')

        if self.enabled_tables.get('variants'):
            self.logger.console.sync.info('\nVariants:\n')

            variants_table = self.variants_tables.get('variants_table')
            self.logger.console.sync.info(f'''{variants_table}\n''')

        if self.mutations_table and self.enabled_tables.get('mutations'):
            self.logger.console.sync.info('\nMutations:\n')
            self.logger.console.sync.info(f'''{self.mutations_table}\n''')

    def _generate_experiments_table(self) -> str:

        experiment_table_rows: List[str] = []  

        header_keys = self.experiment_header_keys.get('experiments_table_headers_keys')
        
        for experiment_summary in self.experiments_summaries.values():

            table_row = OrderedDict()

            for field_name in header_keys:

                experiment_summary_dict = experiment_summary.dict()

                headers = self.experiment_headers.get('experiment_table_headers')
                header_name = headers.get(field_name)
                table_row[header_name] = experiment_summary_dict.get(field_name) 

            experiment_table_rows.append(table_row)


        return tabulate(
            list(sorted(
                experiment_table_rows,
                key=lambda row: row['name']
            )),
            headers='keys',
            missingval='None',
            tablefmt="simple",
            floatfmt='.2f'
        )

    def _generate_variants_tables(self) -> Dict[str, Union[str, None]]:
        
        variants_table_header_keys = self.experiment_header_keys.get('variant_table_headers_keys')
        variants_stats_table_header_keys = self.experiment_header_keys.get('variants_stats_table_header_keys')

        for experiment_summary in self.experiments_summaries.values():

            for variant_summary in experiment_summary.experiment_variant_summaries.values():
                table_row = OrderedDict()

                for field_name in variants_table_header_keys:

                    variant_summary_dict = variant_summary.dict()

                    headers = self.experiment_headers.get('variants_table_headers')
                    header_name = headers.get(field_name)
                    table_row[header_name] = variant_summary_dict.get(field_name) 

                self.variants_table_rows['variants_table_rows'].append(table_row)

                table_row = OrderedDict()

                for field_name in variants_stats_table_header_keys:

                    variant_summary_dict = variant_summary.dict()

                    headers = self.experiment_headers.get('variants_table_headers')
                    header_name = headers.get(field_name)
                    table_row[header_name] = variant_summary_dict.get(field_name) 

                self.variants_table_rows['variant_stats_table_rows'].append(table_row)

        variant_stats_table_rows = list(sorted(
            self.variants_table_rows.get('variant_stats_table_rows'),
            key=lambda row: row['name']
        ))

        for row in variant_stats_table_rows:
            del row['name']

        return {
            'variants_table': tabulate(
                list(sorted(
                    self.variants_table_rows.get('variants_table_rows'),
                    key=lambda row: row['name']
                )),
                headers='keys',
                missingval='None',
                tablefmt="simple",
                floatfmt=".2f"
            ),
            'variant_stats_table': tabulate(
                variant_stats_table_rows,
                headers='keys',
                missingval='None',
                tablefmt="simple",
                floatfmt=".2f"
            )
        }
    
    def _generate_mutations_table(self) -> str:

        mutation_table_rows: List[OrderedDict] = []

        mutations_table_headers_keys = self.experiment_header_keys.get('mutations_table_header_keys')

        for experiment_summary in self.experiments_summaries.values():

            for variant_summary in experiment_summary.experiment_variant_summaries.values():

                for mutation_summary in variant_summary.variant_mutation_summaries.values():

                    mutation_summary_dict = mutation_summary.dict()

                    table_row = OrderedDict()

                    for field_name in mutations_table_headers_keys:

                        headers = self.experiment_headers.get('mutations_table_headers')
                        header_name = headers.get(field_name)
                    
                        table_row[header_name] = mutation_summary_dict.get(field_name) 

                    mutation_table_rows.append(table_row)

        return tabulate(
            list(sorted(
                mutation_table_rows,
                key=lambda row: row['name']
            )),
            headers='keys',
            missingval='None',
            tablefmt="simple"
        )
                


import statistics
from decimal import Decimal
from typing import Dict, List, Optional, Union
from .metrics_set_types import (
    GroupMetricsSet,
    StageMetrics,
    CommonMetrics
)
from hedra.core.personas.streaming.stream_analytics import StreamAnalytics
from .stage_streams_set import StageStreamsSet
from .metrics_set import MetricsSet


class StageMetricsSummary:

    def __init__(
        self,
        stage_name: str=None,
        persona_type: str=None,
        batch_size: int=None,
        total_elapsed: float=None,
        stage_streamed_analytics: Optional[List[StreamAnalytics]]=None
    ) -> None:

        self.metrics_sets: Dict[str, MetricsSet] = {}
        self.action_and_task_metrics: Dict[str, GroupMetricsSet] = {}

        self.stage_streamed_analytics = stage_streamed_analytics

        initial_stream = []
        if self.stage_streamed_analytics:
            initial_stream = [
                0 for _ in range(len(
                    self.stage_streamed_analytics[0].interval_completed_counts
                ))
            ]

        persona_type_slug = '-'.join([
            segment.capitalize() for segment in persona_type.split('-')
        ])

        self.stage_metrics = StageMetrics(**{
            'name': stage_name,
            'persona': persona_type_slug,
            'batch_size': batch_size,
            'total': 0,
            'succeeded': 0,
            'failed': 0,
            'aps': 0.,
            'time': total_elapsed,
            'μ_sec': 0.,
            'μ_waiting': 0.,
            'μ_connecting': 0.,
            'μ_reading': 0.,
            'μ_writing': 0.,
            'streamed_completion_rates': list(initial_stream),
            'streamed_completed': list(initial_stream),
            'streamed_succeeded': list(initial_stream),
            'streamed_failed': list(initial_stream),
            'streamed_batch_timings': list(initial_stream),
        })

        self.stage_table_header_keys = [
            'name',
            'time',
            'persona',
            'batch_size',
            'total',
            'succeeded',
            'failed',
            'aps',
            'μ_sec',
            'μ_waiting',
            'μ_connecting',
            'μ_reading',
            'μ_writing'
        ]

        self.stage_table_headers = {
            header_name: header_name.replace(
                '_', ' '
            ) for header_name in self.stage_table_header_keys
        }

        self.groups: List[str] = [
            'total',
            'waiting',
            'connecting',
            'writing',
            'reading'
        ]

        self.common_fields: List[str] = [
            'total',
            'succeeded',
            'failed'
        ]

        self.fields: List[str] = [
            'med',
            'μ',
            'var',
            'std',
            'min',
            'max',
            'q_10',
            'q_20',
            'q_30',
            'q_40',
            'q_50',
            'q_60',
            'q_70',
            'q_80',
            'q_90',
            'q_95',
            'q_99'
        ]

        self._record_fields: List[str] = [
            'median',
            'mean',
            'variance',
            'stdev',
            'minimum',
            'maximum',
            'quantile_10th',
            'quantile_20th',
            'quantile_30th',
            'quantile_40th',
            'quantile_50th',
            'quantile_60th',
            'quantile_70th',
            'quantile_80th',
            'quantile_90th',
            'quantile_95th',
            'quantile_99th'
        ]

        self._record_fields_map: Dict[str, str] = {}

        for record_field, table_field in zip(self._record_fields, self.fields):
            self._record_fields_map[record_field] = table_field
        
        self.action_or_task_header_keys: List[str] = []

        for group in self.groups:
            for field in self.fields:
                self.action_or_task_header_keys.append(
                    f'{group}_{field}'
                )

        self.action_or_task_headers = {
            header_name: header_name.replace(
                '_', ' '
            ) for header_name in self.action_or_task_header_keys
        }

        self.common_metrics: Dict[str, CommonMetrics] = {}

    @property
    def streams(self):
        return StageStreamsSet(
            self.stage_metrics.name,
            self.stage_streamed_analytics
        )

    def calculate_action_and_task_metrics(self):
        self._group_action_and_task_metrics()
        self._calculate_stage_totals()

        if self.stage_streamed_analytics:
            self._calculate_completed_rate_stream()
            self._calculate_succeeded_rate_stream()
            self._calculate_failed_rate_stream()
            self._calculate_batch_timings_stream()
            self._calculate_completion_rate_stream()

    def _group_action_and_task_metrics(self):

        for action_or_task_name, metric_set in self.metrics_sets.items():

            self.stage_metrics.total += metric_set.common_stats.get('total')
            self.stage_metrics.succeeded += metric_set.common_stats.get('succeeded')
            self.stage_metrics.failed += metric_set.common_stats.get('failed')

            grouped_metrics = {}
            common_metrics = {}

            for group_name, group_metrics in metric_set.groups.items():

                for field_name, field_value in group_metrics.record.items():
                    if isinstance(field_value, (int, float,)):

                        table_field_name = self._record_fields_map.get(field_name)
                        metric_key = f'{group_name}_{table_field_name}'

                        grouped_metrics[metric_key] = field_value

            for table_key in self.common_fields:
                common_metrics[table_key] = metric_set.common_stats.get(table_key)
            
            common_metrics['aps'] = metric_set.common_stats.get('actions_per_second')

            self.common_metrics[action_or_task_name] = CommonMetrics(**common_metrics)
            self.action_and_task_metrics[action_or_task_name] = GroupMetricsSet(**grouped_metrics)
        
    def _calculate_stage_totals(self):
        self.stage_metrics.aps = round(
            self.stage_metrics.total/self.stage_metrics.time,
            2
        )

        for group_metrics_set in self.action_and_task_metrics.values():
            self.stage_metrics.μ_sec = group_metrics_set.total_μ
            self.stage_metrics.μ_waiting += group_metrics_set.waiting_μ
            self.stage_metrics.μ_connecting += group_metrics_set.connecting_μ
            self.stage_metrics.μ_writing += group_metrics_set.writing_μ
            self.stage_metrics.μ_reading += group_metrics_set.reading_μ


        self.stage_metrics.μ_sec = Decimal(
            self.stage_metrics.μ_sec/self.stage_metrics.total
        )

        self.stage_metrics.μ_waiting = Decimal(
            self.stage_metrics.μ_waiting/self.stage_metrics.total
        )

        self.stage_metrics.μ_connecting = Decimal(
            self.stage_metrics.μ_connecting/self.stage_metrics.total
        )

        self.stage_metrics.μ_reading = Decimal(
            self.stage_metrics.μ_reading/self.stage_metrics.total
        )

        self.stage_metrics.μ_writing = Decimal(
            self.stage_metrics.μ_writing/self.stage_metrics.total
        )

    def _calculate_completion_rate_stream(self):
        bins_count = max(
            len(stream.interval_completion_rates) for stream in self.stage_streamed_analytics
        )
        self.stage_metrics.streamed_completion_rates = [0 for _ in range(bins_count)]
        
        for idx in range(bins_count):

            self.stage_metrics.streamed_completion_rates[idx] = round(
                self.stage_metrics.streamed_completed[idx]/self.stage_metrics.streamed_batch_timings[idx],
                2
            )

    def _calculate_succeeded_rate_stream(self):
        bins_count = max(
            len(stream.interval_succeeded_counts) for stream in self.stage_streamed_analytics
        )

        self.stage_metrics.streamed_succeeded = [0 for _ in range(bins_count)]

        for stream in self.stage_streamed_analytics:
            for idx, succeeded_count in enumerate(stream.interval_succeeded_counts):
                self.stage_metrics.streamed_succeeded[idx] += succeeded_count        

        total_succeeded = sum(self.stage_metrics.streamed_succeeded) 

        if total_succeeded > self.stage_metrics.succeeded:
            stream_error = total_succeeded - self.stage_metrics.succeeded
            binned_error = int(stream_error/bins_count)
            remainder_error = stream_error%bins_count

            carried = 0
            for idx in range(bins_count):
                error = binned_error + carried
                corrected_amount = self.stage_metrics.streamed_succeeded[idx] - error

                if corrected_amount > 0:
                    self.stage_metrics.streamed_succeeded[idx] = corrected_amount
                    carried = 0

                else:
                    carried += error

            self.stage_metrics.streamed_succeeded[-1] -= remainder_error

    def _calculate_failed_rate_stream(self):
        bins_count = max(
            len(stream.interval_failed_counts) for stream in self.stage_streamed_analytics
        )

        self.stage_metrics.streamed_failed = [0 for _ in range(bins_count)]

        for stream in self.stage_streamed_analytics:
            for idx, failed_count in enumerate(stream.interval_failed_counts):
                self.stage_metrics.streamed_failed[idx] += failed_count

        total_failed = sum(self.stage_metrics.streamed_failed) 

        if total_failed > self.stage_metrics.failed:
            stream_error = total_failed - self.stage_metrics.failed
            binned_error = int(stream_error/bins_count)
            remainder_error = stream_error%bins_count

            carried = 0
            for idx in range(bins_count):
                error = binned_error + carried
                corrected_amount = self.stage_metrics.streamed_failed[idx] - error

                if corrected_amount > 0:
                    self.stage_metrics.streamed_failed[idx] = corrected_amount
                    carried = 0

                else:
                    carried += error

            self.stage_metrics.streamed_failed[-1] -= remainder_error

    def _calculate_completed_rate_stream(self):
        bins_count = max([
            len(stream.interval_completed_counts) for stream in self.stage_streamed_analytics
        ])
        self.stage_metrics.streamed_completed = [0 for _ in range(bins_count)]

        for stream in self.stage_streamed_analytics:
            for idx, completed_count in enumerate(stream.interval_completed_counts):
                self.stage_metrics.streamed_completed[idx] += completed_count

        total_completed = sum(self.stage_metrics.streamed_completed) 

        if total_completed > self.stage_metrics.total:
            stream_error = total_completed - self.stage_metrics.total
            binned_error = int(stream_error/bins_count)
            remainder_error = stream_error%bins_count

            carried = 0
            for idx in range(bins_count):
                error = binned_error + carried
                corrected_amount = self.stage_metrics.streamed_completed[idx] - error

                if corrected_amount > 0:
                    self.stage_metrics.streamed_completed[idx] = corrected_amount
                    carried = 0

                else:
                    carried += error

            self.stage_metrics.streamed_completed[-1] -= remainder_error

    def _calculate_batch_timings_stream(self):
        bins_count = max([
            len(stream.interval_batch_timings) for stream in self.stage_streamed_analytics
        ])

        streamed_batch_timings = [[] for _ in range(bins_count)]

        for stream in self.stage_streamed_analytics:
            for idx, batch_timing in enumerate(stream.interval_batch_timings):
                streamed_batch_timings[idx].append(batch_timing)

        for idx, batch_timings in enumerate(streamed_batch_timings):
            streamed_batch_timings[idx] = statistics.median(batch_timings)

        for idx, batch_timing in enumerate(streamed_batch_timings):
            streamed_batch_timings[idx] = round(
                batch_timing,
                2
            )

        self.stage_metrics.streamed_batch_timings = streamed_batch_timings



from hedra.distributed.snowflake import Snowflake
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt
)
from typing import Any, Dict, Union

class Entry(BaseModel):
    entry_id: StrictInt
    key: StrictStr
    value: Any
    term: StrictInt
    leader_host: StrictStr
    leader_port: StrictInt
    timestamp: StrictInt

    def __init__(self, *args, **kwargs):

        entry_id: Union[int, None] = kwargs.get('entry_id')
        if entry_id:
            kwargs['timestamp'] = Snowflake.parse(entry_id).timestamp

        super().__init__(
            *args,
            **kwargs
        )

    def to_data(self):
        return {
            'key': self.key,
            'value': self.value,
            'timestamp': self.timestamp
        }
    
    @classmethod
    def from_data(
        cls,
        entry_id: int,
        leader_host: str,
        leader_port: int,
        term: int,
        data: Dict[str, Any]
    ):

        return Entry(
            entry_id=entry_id,
            leader_host=leader_host,
            leader_port=leader_port,
            term=term,
            **data
        )


from typing import (
    Coroutine, 
    Dict, 
    List, 
    Tuple, 
    Union
)
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.task import MercuryTaskRunner, Task, TaskResult
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.types.tracing.trace_session import TraceSession
from hedra.logging import HedraLogger
from .base_client import BaseClient


class TaskClient(BaseClient[MercuryTaskRunner, Task, TaskResult]):

    def __init__(self, config: Config) -> None:
        super().__init__()

        if config is None:
            config = Config()
        
        tracing_session: Union[TraceSession, None] = None
        if config.tracing:
            trace_config_dict = config.tracing.to_dict()
            tracing_session = TraceSession(**trace_config_dict)

        self.session = MercuryTaskRunner(
            concurrency=config.batch_size,
            timeouts=Timeouts(
                total_timeout=config.request_timeout
            ),
            tracing_session=tracing_session
        )
        self.request_type = RequestTypes.TASK
        self.client_type = self.request_type.capitalize()
        self.next_name = None

        self.logger = HedraLogger()
        self.logger.initialize()

    def call(
        self, 
        task: Coroutine, 
        env: str = None,
        user: str = None,
        tags: List[Dict[str, str]] = []
    ) -> Tuple[Task, MercuryTaskRunner]:
        task_action = Task(
            self.next_name,
            task,
            source=env,
            user=user,
            tags=tags
        )

        return task_action, self.session


import psutil
from collections import defaultdict
from typing import Dict, List, Union, Any
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.engines.types.graphql import GraphQLResult
from hedra.core.engines.types.graphql_http2 import GraphQLHTTP2Result
from hedra.core.engines.types.grpc import GRPCResult
from hedra.core.engines.types.http import HTTPResult
from hedra.core.engines.types.http2 import HTTP2Result
from hedra.core.engines.types.playwright import PlaywrightResult
from hedra.core.engines.types.task import TaskResult
from hedra.core.engines.types.udp import UDPResult
from hedra.core.engines.types.websocket import WebsocketResult
from hedra.core.personas.streaming.stream_analytics import StreamAnalytics
from hedra.data.serializers import Serializer
from hedra.monitoring import (
    CPUMonitor,
    MemoryMonitor
)


ResultsBatch = Dict[str, Union[List[BaseResult], float]]

MutationConfig = Dict[str, Union[str, float, List[str]]]

VariantConfig = Dict[str, Union[int, str, float, List[float], MutationConfig]]

ExperimentConfig = Dict[str, Union[str, bool, VariantConfig]]

MemoryMonitorGroup = Dict[str, MemoryMonitor]

CPUMonitorGroup = Dict[str, CPUMonitor]

MonitorGroup = Dict[str, Union[CPUMonitorGroup, MemoryMonitorGroup]]


class ResultsSet:

    def __init__(
        self,
        execution_results: Dict[str, Union[int, float, List[ResultsBatch], ExperimentConfig]]
    ) -> None:

        self.stage: str = execution_results.get('stage')
        self.stage_streamed_analytics: Union[List[StreamAnalytics], None] = execution_results.get('streamed_analytics')

        self.total_elapsed: float = execution_results.get('total_elapsed', 0)
        self.total_results: int = execution_results.get('total_results', 0)

        self.stage_batch_size = execution_results.get('stage_batch_size', 0)
        self.stage_optimized = execution_results.get('stage_optimized', False)
        self.stage_persona_type = execution_results.get('stage_persona_type', 'default')
        self.stage_workers = execution_results.get(
            'stage_workers',
            psutil.cpu_count(logical=False)
        )

        self.results: List[BaseResult] = execution_results.get('stage_results', [])

        self.serialized_results: List[Dict[str, Any]] = execution_results.get('serialized_results', [])
        self.experiment = execution_results.get('experiment')
        self.serializer = Serializer()

        self.types = {
            RequestTypes.GRAPHQL: GraphQLResult,
            RequestTypes.GRAPHQL_HTTP2: GraphQLHTTP2Result,
            RequestTypes.GRPC: GRPCResult,
            RequestTypes.HTTP: HTTPResult,
            RequestTypes.HTTP2: HTTP2Result,
            RequestTypes.PLAYWRIGHT: PlaywrightResult,
            RequestTypes.TASK: TaskResult,
            RequestTypes.UDP: UDPResult,
            RequestTypes.WEBSOCKET: WebsocketResult
        }

    def __iter__(self):
        for result in self.results:
            yield result


    def to_serializable(self):
        return {
            'total_elapsed': self.total_elapsed,
            'total_results': self.total_results,
            'results': [
                self.serializer.serialize_result(result) for result in self.results
            ]
        }
    
    def load_results(self):
        self.results = [
            self.serializer.deserialize_result(result) for result in self.serialized_results
        ]

    def group(self) -> Dict[str, List[BaseResult]]:
        grouped_results = defaultdict(list)
        for result in self.results:
            grouped_results[result.name].append(result)

        return grouped_results
    
    def copy(self):
        return ResultsSet({
            'stage': self.stage,
            'streamed_analytics': list(self.stage_streamed_analytics),
            'stage_batch_size': self.stage_batch_size,
            'stage_persona_type': self.stage_persona_type,
            'stage_workers': self.stage_workers,
            'stage_optimized': self.stage_optimized,
            'total_elapsed': self.total_elapsed,
            'total_results': self.total_results,
            'stage_results': list(self.results),
            'serialized_results': list(self.serialized_results)
        })


import psutil
import itertools
import statistics
from hedra.monitoring.base.monitor import BaseMonitor
from typing import Optional


class CPUMonitor(BaseMonitor):

    def __init__(self) -> None:
        super().__init__()

    def update_monitor(self, monitor_name: str):
        self.active[monitor_name].append(
            psutil.cpu_percent()
        )

    def aggregate_worker_stats(self):
        
        monitor_stats = self._collect_worker_stats()

        for monitor_name, metrics in monitor_stats.items():
            self.stage_metrics[monitor_name] = [
                statistics.median(cpu_usage) for cpu_usage in itertools.zip_longest(
                    *metrics,
                    fillvalue=0
                )
            ]

        


from typing import Any, Dict, List, Union
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.grpc import (
    MercuryGRPCClient,
    GRPCAction,
    GRPCResult
)
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.client.store import ActionsStore
from hedra.core.engines.types.tracing.trace_session import (
    TraceSession, 
    Trace
)
from hedra.logging import HedraLogger
from .base_client import BaseClient


class GRPCClient(BaseClient[MercuryGRPCClient, GRPCAction, GRPCResult]):
    
    def __init__(self, config: Config) -> None:
        super().__init__()

        if config is None:
            config = Config()

        tracing_session: Union[TraceSession, None] = None
        if config.tracing:
            trace_config_dict = config.tracing.to_dict()
            tracing_session = TraceSession(**trace_config_dict)

        self.session = MercuryGRPCClient(
            concurrency=config.batch_size,
            timeouts=Timeouts(
                total_timeout=config.request_timeout
            ),
            reset_connections=config.reset_connections,
            tracing_session=tracing_session
        )
        self.request_type = RequestTypes.GRPC
        self.client_type = self.request_type.capitalize()

        self.actions: ActionsStore = None
        self.next_name = None
        self.intercept = False

        self.logger = HedraLogger()
        self.logger.initialize()

    def __getitem__(self, key: str):
        return self.session.registered.get(key)

    async def request(
        self, 
        url: str, 
        headers: Dict[str, str] = {}, 
        protobuf: Any = None, 
        user: str = None, 
        tags: List[Dict[str, str]] = [],
        trace: Trace=None
    ):
        if trace and self.session.tracing_session is None:
            self.session.tracing_session = TraceSession(
                **trace.to_dict()
            )

        request = GRPCAction(
            self.next_name,
            url,
            method='POST',
            headers=headers,
            data=protobuf,
            user=user,
            tags=tags
        )

        return await self._execute_action(request)


import uuid
import numpy
import statistics
from collections import defaultdict
from typing import Dict, List, Union
from .system_metrics_group import SystemMetricsGroup
from .system_metrics_set_types import (
    MonitorGroup,
    MemoryMonitorGroup,
    CPUMonitorGroup,
    SystemMetricsCollection,
    SessionMetricsCollection,
    SystemMetricGroupType
)


class SystemMetricsSet:

    metrics_table_keys = [
        'name',
        'mean',
        'median',
        'max',
        'min',
        'stdev',
        'variance'
    ]

    session_metrics_metadata = [
        'name',
        'group',
    ]

    stage_metrics_metadata = [
        'stage',
        'name',
        'group',
    ]

    primary_metrics = [
        'mean',
        'median',
        'max',
        'min',
        'stdev',
        'variance',
    ]

    quantiles = [
        'quantile_10th',
        'quantile_20th',
        'quantile_25th',
        'quantile_30th',
        'quantile_40th',
        'quantile_50th',
        'quantile_60th',
        'quantile_70th',
        'quantile_75th',
        'quantile_80th',
        'quantile_90th',
        'quantile_95th',
        'quantile_99th'
    ]

    metrics_header_keys = [
        'stage',
        'name',
        'group',
        'mean',
        'median',
        'max',
        'min',
        'stdev',
        'variance',
        'quantile_10th',
        'quantile_20th',
        'quantile_25th',
        'quantile_30th',
        'quantile_40th',
        'quantile_50th',
        'quantile_60th',
        'quantile_70th',
        'quantile_75th',
        'quantile_80th',
        'quantile_90th',
        'quantile_95th',
        'quantile_99th'
    ]

    def __init__(
            self, 
            metrics: MonitorGroup,
            batch_sizes: Dict[str, int]
        ) -> None:
        
        self.system_metrics_set_id = uuid.uuid4()
        self.system_cpu_metrics: Dict[str, List[Union[int, float]]] = defaultdict(list)
        self.system_memory_metrics: Dict[str, List[Union[int, float]]] = defaultdict(list)

        self.session_cpu_metrics: Dict[str, SystemMetricsCollection] = {}
        self.session_memory_metrics: Dict[str, SystemMetricsCollection] = {}
        self.mb_per_vu: Dict[str, SystemMetricsCollection] = {}
        self.batch_sizes = batch_sizes

        self._quantiles = [
            10,
            20,
            25,
            30,
            40,
            50,
            60,
            70,
            75,
            80,
            90,
            95,
            99
        ]

        self.metrics = metrics

        self.cpu_metrics_by_stage: Dict[str, CPUMonitorGroup] = {
            stage_name: stage_metrics.get(
                'cpu'
            ) for stage_name, stage_metrics in metrics.items()
        }

        self.memory_metrics_by_stage: Dict[str, MemoryMonitorGroup] = {
            stage_name: stage_metrics.get(
                'memory'
            ) for stage_name, stage_metrics in metrics.items()
        }

        self.cpu = SystemMetricsGroup(
            self.cpu_metrics_by_stage,
            SystemMetricGroupType.CPU
        )

        self.memory = SystemMetricsGroup(
            self.memory_metrics_by_stage,
            SystemMetricGroupType.MEMORY
        )

    def generate_system_summaries(self):
        self.cpu.aggregate()
        self.memory.aggregate()

        for stage_metrics in self.metrics.values():
            cpu_metrics_group = stage_metrics.get('cpu')

            for monitor_name, monitor_metrics in cpu_metrics_group.collected.items():
                self.system_cpu_metrics[monitor_name].extend(monitor_metrics)
        
        for stage_name, stage_metrics in self.metrics.items():
            memory_metrics_group = stage_metrics.get('memory')

            for monitor_name, monitor_metrics in memory_metrics_group.collected.items():
                self.system_memory_metrics[monitor_name].extend(monitor_metrics)

                is_visible = memory_metrics_group.visibility_filters[monitor_name]
                stage_batch_size = self.batch_sizes.get(stage_name)

                if memory_metrics_group.is_execute_stage and is_visible and stage_batch_size:


                    mb_per_vu = [
                        round(
                            memory_used/(1024**2 * stage_batch_size),
                            2
                        ) for memory_used in monitor_metrics
                    ]

                    self.mb_per_vu[stage_name] = SystemMetricsCollection(**{
                        'stage': stage_name,
                        'name': monitor_name,
                        'group': SystemMetricGroupType.MEMORY.value,
                        'mean': statistics.mean(mb_per_vu),
                        'median': statistics.median(mb_per_vu),
                        'max': max(mb_per_vu),
                        'min': min(mb_per_vu),
                        'stdev': statistics.stdev(mb_per_vu),
                        'variance': statistics.variance(mb_per_vu),
                        **{
                            f'quantile_{quantile}th':  numpy.quantile(
                                mb_per_vu,
                                round(
                                    quantile/100,
                                    2
                                )
                            ) for quantile in self._quantiles
                        }
                    })


        for monitor_name, monitor_metrics in self.system_cpu_metrics.items():
            self.session_cpu_metrics[monitor_name] = SessionMetricsCollection(**{
                'name': monitor_name,
                'group': SystemMetricGroupType.CPU.value,
                'mean': statistics.mean(monitor_metrics),
                'median': statistics.median(monitor_metrics),
                'max': max(monitor_metrics),
                'min': min(monitor_metrics),
                'stdev': statistics.stdev(monitor_metrics),
                'variance': statistics.variance(monitor_metrics),
                **{
                    f'quantile_{quantile}th':  numpy.quantile(
                        monitor_metrics,
                        round(
                            quantile/100,
                            2
                        )
                    ) for quantile in self._quantiles
                }
            })

        for monitor_name, monitor_metrics in self.system_memory_metrics.items():
            self.session_memory_metrics[monitor_name] = SessionMetricsCollection(**{
                'name': monitor_name,
                'group': SystemMetricGroupType.MEMORY.value,
                'mean': statistics.mean(monitor_metrics),
                'median': statistics.median(monitor_metrics),
                'max': max(monitor_metrics),
                'min': min(monitor_metrics),
                'stdev': statistics.stdev(monitor_metrics),
                'variance': statistics.variance(monitor_metrics),
                **{
                    f'quantile_{quantile}th':  numpy.quantile(
                        monitor_metrics,
                        round(
                            quantile/100,
                            2
                        )
                    ) for quantile in self._quantiles
                }
            })



from hedra.reporting.types.common.types import ReporterTypes
from pydantic import BaseModel


class InfluxDBConfig(BaseModel):
    host: str='localhost:8086'
    token: str
    organization: str='hedra'
    connect_timeout: int=10000
    events_bucket: str='events'
    metrics_bucket: str='metrics'
    experiments_bucket: str='experiments'
    streams_bucket: str='streams'
    system_metrics_bucket: str='system_metrics'
    secure: bool=False
    reporter_type: ReporterTypes=ReporterTypes.InfluxDB


import uuid
import json
from typing import List, Dict, Any
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from .kafka_config import KafkaConfig

try:

    from aiokafka import AIOKafkaProducer
    has_connector = True

except Exception:
    AIOKafkaProducer = None
    has_connector = False


class Kafka:

    def __init__(self, config: KafkaConfig) -> None:
        self.host = config.host
        self.client_id = config.client_id

        self.events_topic = config.events_topic
        self.metrics_topic = config.metrics_topic
        self.streams_topic = config.streams_topic

        self.custom_metrics_topic = f'{config.metrics_topic}_custom' 
        self.shared_metrics_topic = f'{config.metrics_topic}_shared'
        self.errors_topic = f'{config.metrics_topic}_errors'

        self.experiments_topic = config.experiments_topic
        self.variants_topic = f'{config.experiments_topic}_variants'
        self.mutations_topic = f'{config.experiments_topic}_metrics'

        self.session_system_metrics_topic= f'{config.system_metrics_topic}_session'
        self.stage_system_metrics_topic = f'{config.system_metrics_topic}_stage'

        self.events_partition = config.events_partition
        self.metrics_partition = config.metrics_partition
        self.experiments_partition = config.experiments_partition
        self.streams_partition = config.streams_partition

        self.system_metrics_partition = config.system_metrics_partition

        self.shared_metrics_partition = config.metrics_partition
        self.errors_partition = config.metrics_partition
        self.custom_metrics_partition = config.metrics_partition

        self.compression_type = config.compression_type
        self.timeout = config.timeout
        self.enable_idempotence = config.idempotent or True
        self.options: Dict[str, Any] = config.options or {}
        self._producer = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Kafka at - {self.host}')

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - Compression Type: {self.compression_type}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - Connection Timeout: {self.timeout}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - Idempotent: {self.enable_idempotence}')

        for option_name, option in self.options.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - {option_name.capitalize()}: {option}')


        self._producer = AIOKafkaProducer(
            bootstrap_servers=self.host,
            client_id=self.client_id,
            compression_type=self.compression_type,
            request_timeout_ms=self.timeout,
            enable_idempotence=self.enable_idempotence,
            **self.options
        )

        await self._producer.start()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Kafka at - {self.host}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metric to Topic - {self.session_system_metrics_topic} - Partition - {self.system_metrics_partition}')
        
        batch = self._producer.create_batch()

        metrics_sets: List[SessionMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.system_metrics_set_id}')
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics)

        for metric_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metric - {metric_set.name}:{metric_set.group}')

            batch.append(
                value=json.dumps(
                    metric_set.record
                ).encode('utf-8'),
                timestamp=None, 
                key=bytes(f'{metric_set.name}_{metric_set.group}', 'utf')
            )

        await self._producer.send_batch(
            batch,
            self.session_system_metrics_topic,
            partition=self.system_metrics_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metric to Topic - {self.session_system_metrics_topic} - Partition - {self.system_metrics_partition}')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metric to Topic - {self.stage_system_metrics_topic} - Partition - {self.system_metrics_partition}')
        
        batch = self._producer.create_batch()

        metrics_sets: List[SystemMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.system_metrics_set_id}')
            
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics)

        for metric_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metric - {metric_set.name}:{metric_set.group}')

            batch.append(
                value=json.dumps(
                    metric_set.record
                ).encode('utf-8'),
                timestamp=None, 
                key=bytes(f'{metric_set.name}_{metric_set.group}', 'utf')
            )

        await self._producer.send_batch(
            batch,
            self.stage_system_metrics_topic,
            partition=self.system_metrics_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metric to Topic - {self.stage_system_metrics_topic} - Partition - {self.system_metrics_partition}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Topic - {self.streams_topic} - Partition - {self.streams_partition}')
        
        batch = self._producer.create_batch()
        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams - {stage_name}:{stream.stream_set_id}')

            for group_name, group in stream.grouped.items():
                batch.append(
                    value=json.dumps(
                        {
                            **group,
                            'name': f'{stage_name}_streams',
                            'stage': stage_name,
                            'group': group_name
                        }
                    ).encode('utf-8'),
                    timestamp=None, 
                    key=bytes(f'{stage_name}_{group_name}', 'utf')
                )

        await self._producer.send_batch(
            batch,
            self.streams_topic,
            partition=self.streams_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to Topic - {self.streams_topic} - Partition - {self.streams_partition}')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Topic - {self.experiments_topic} - Partition - {self.experiments_partition}')

        batch = self._producer.create_batch()
        for experiment in experiment_metrics.experiment_summaries:

            batch.append(
                value=json.dumps(
                    experiment.record
                ).encode('utf-8'),
                timestamp=None, 
                key=bytes(experiment.experiment_name, 'utf')
            )

        await self._producer.send_batch(
            batch,
            self.experiments_topic,
            partition=self.experiments_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to Topic - {self.experiments_topic} - Partition - {self.experiments_partition}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Topic - {self.variants_topic} - Partition - {self.experiments_partition}')

        batch = self._producer.create_batch()
        for variant in experiment_metrics.variant_summaries:

            batch.append(
                value=json.dumps(
                    variant.record
                ).encode('utf-8'),
                timestamp=None, 
                key=bytes(variant.variant_name, 'utf')
            )

        await self._producer.send_batch(
            batch,
            self.variants_topic,
            partition=self.experiments_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to Topic - {self.variants_topic} - Partition - {self.experiments_partition}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to Topic - {self.mutations_topic} - Partition - {self.experiments_partition}')

        batch = self._producer.create_batch()
        for mutation in experiment_metrics.mutation_summaries:

            batch.append(
                value=json.dumps(
                    mutation.record
                ).encode('utf-8'),
                timestamp=None, 
                key=bytes(mutation.mutation_name, 'utf')
            )

        await self._producer.send_batch(
            batch,
            self.mutations_topic,
            partition=self.events_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to Topic - {self.mutations_topic} - Partition - {self.experiments_partition}')

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Topic - {self.events_topic} - Partition - {self.events_partition}')

        batch = self._producer.create_batch()
        for event in events:

            batch.append(
                value=json.dumps(
                    event.record
                ).encode('utf-8'),
                timestamp=None, 
                key=bytes(event.name, 'utf')
            )

        await self._producer.send_batch(
            batch,
            self.events_topic,
            partition=self.events_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Topic - {self.events_topic} - Partition - {self.events_partition}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Topic - {self.shared_metrics_topic} - Partition - {self.shared_metrics_partition}')

        batch = self._producer.create_batch()
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            batch.append(
                value=json.dumps({
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': 'common',
                    **metrics_set.common_stats
                }).encode('utf-8'),
                timestamp=None, 
                key=bytes(metrics_set.name, 'utf')
            )

        await self._producer.send_batch(
            batch,
            self.shared_metrics_topic,
            partition=self.shared_metrics_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Topic - {self.shared_metrics_topic} - Partition - {self.shared_metrics_partition}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Topic - {self.metrics_topic} - Partition - {self.metrics_partition}')
        
        batch = self._producer.create_batch()
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                batch.append(
                    value=json.dumps(
                        {
                            **group.record,
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': group_name
                        }
                    ).encode('utf-8'),
                    timestamp=None, 
                    key=bytes(f'{metrics_set.name}_{group_name}', 'utf')
                )

        await self._producer.send_batch(
            batch,
            self.metrics_topic,
            partition=self.metrics_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Topic - {self.metrics_topic} - Partition - {self.metrics_partition}')
    
    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        batch = self._producer.create_batch()

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Customm Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_metic_name, custom_metric in metrics_set.custom_metrics.items():

                batch.append(
                    value=json.dumps({
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': 'custom',
                        custom_metic_name: custom_metric.metric_value
                    }).encode('utf-8'),
                    timestamp=None,
                    key=bytes(f'{metrics_set.name}_{custom_metic_name}', 'utf')
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Customm Metrics to Topic - {self.custom_metrics_topic} - Partition - {self.custom_metrics_partition}')
        await self._producer.send_batch(
            batch,
            self.custom_metrics_topic,
            partition=self.custom_metrics_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Customm Metrics to Topic - {self.custom_metrics_topic} - Partition - {self.custom_metrics_partition}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Topic - {self.metrics_topic} - Partition - {self.metrics_partition}')

        batch = self._producer.create_batch()
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                batch.append(
                    value=json.dumps(
                        {
                            'metric_name': metrics_set.name,
                            'metric_stage': metrics_set.stage,
                            'error_message': error.get('message'),
                            'error_count': error.get('count')
                        }
                    ).encode('utf-8'),
                    timestamp=None, 
                    key=bytes(metrics_set.name, 'utf')
                )
        
        await self._producer.send_batch(
            batch,
            self.errors_topic,
            partition=self.errors_partition
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Topic - {self.metrics_topic} - Partition - {self.metrics_partition}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connection to Kafka at - {self.host}')

        await self._producer.stop()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connection to Kafka at - {self.host}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')


from __future__ import annotations
import inspect
from typing import Any, Generic, TypeVar
from hedra.core.engines.types.common.base_action import BaseAction
from typing import Dict, List
from hedra.core.engines.types.common.types import RequestTypes

A = TypeVar('A')

class Action(BaseAction[A], Generic[A]):

    __slots__ = (
        "source",
        "use_security_context",
        "security_context",
        "type"
    )


    def __init__(self, 
        name: str=None,
        source: str=None,
        user: str=None, 
        tags: List[Dict[str, str]] = [],
        use_security_context: bool = False
    ) -> Any:

        self.source = source
        self.use_security_context = use_security_context
        self.security_context: Any = None
        self.type = A
        self.plugin_type = None

        super().__init__(
            name,
            user,
            tags
        )

    def setup(self):
        pass

    def action_to_serializable(self):
        attributes = inspect.getmembers(
            self, 
            lambda attr: not(inspect.isroutine(attr))
        )

        instance_attributes = [
            attr for attr in attributes if not(
                attr[0].startswith('__') and attr[0].endswith('__')
            )
        ]

        base_attributes = set(dir(Action) + dir(BaseAction))

        serializable = {
            'name': self.name,
            'type': str(self.type),
            'plugin_type': self.plugin_type,
            'use_security_context': self.use_security_context,
            'security_context': self.security_context,
            'metadata': {
                'user': self.metadata.user,
                'tags': self.metadata.tags
            },
            'hooks': self.hooks.to_names(),
            'fields': {}
        }
        for attribute_name, attribute_value in instance_attributes:
            
            if serializable.get(attribute_name) is None and attribute_name not in base_attributes:
                serializable['fields'][attribute_name] = attribute_value

        return serializable


from pydantic import (
    BaseModel, 
    Field, 
    StrictStr,
    StrictBool, 
    validator
)


class MetricHookValidator(BaseModel):
    metric_type: StrictStr
    group: StrictStr=Field(..., min_length=1)
    skip: StrictBool

    @validator('metric_type')
    def validate_names(cls, val):
        
        valid_metric_types = [
            "count",
            "rate",
            "distribution",
            "sample"
        ]

        assert val in valid_metric_types

        return val




from typing import Tuple, Optional
from pydantic import (
    BaseModel, 
    validator, 
    StrictStr, 
    StrictInt,
    StrictBool
)


class TransformHookValidator(BaseModel):
    names: Optional[Tuple[StrictStr, ...]]
    order: Optional[StrictInt]
    skip: StrictBool

    @validator('names')
    def validate_names(cls, vals):
        assert len(vals) == len(set(vals))

        return vals



from hedra.distributed.env.memory_parser import MemoryParser
from hedra.distributed.env.time_parser import TimeParser
from pydantic import (
    BaseModel,
    IPvAnyAddress,
    StrictStr,
    StrictInt,
    StrictBool,
    StrictFloat
)
from typing import (
    Optional, 
    Callable, 
    Literal,
    List,
    Union
)

from .request import Request


HTTPMethod = Literal[
    "GET",
    "HEAD",
    "OPTIONS",
    "POST",
    "PUT",
    "PATCH",
    "DELETE",
    "TRACE"
]


class Limit(BaseModel):
    max_requests: StrictInt
    min_requests: Optional[StrictInt]
    request_period: StrictStr='1s'
    reject_requests: StrictBool=True
    request_backoff: StrictStr='1s'
    cpu_limit: Optional[Union[StrictFloat, StrictInt]]
    memory_limit: Optional[StrictStr]
    limiter_type: Optional[
        Literal[
            "adaptive",
            "cpu-adaptive",
            "leaky-bucket",
            "rate-adaptive",
            "sliding-window",
            "token-bucket",
        ]
    ]
    limit_key: Optional[
        Callable[
            [
                Request,
                IPvAnyAddress,
            ],
            str
        ]
    ]
    rules: Optional[
        List[
            Callable[
                [
                    Request,
                    IPvAnyAddress,

                ],
                bool
            ]
        ]
    ]

    @property
    def backoff(self):
        return TimeParser(self.request_backoff).time

    @property
    def period(self):
        return TimeParser(self.request_period).time
    
    @property
    def memory(self):
        return MemoryParser(self.memory_limit).megabytes(accuracy=4)

    def get_key(
        self, 
        request: Request,
        ip_address: IPvAnyAddress,
        default: str='default'
    ):

        if self.limit_key is None:
            return default
        
        return self.limit_key(
            request,
            ip_address
        )

    def matches(
        self,
        request: Request,
        ip_address: IPvAnyAddress
    ):

        if self.rules is None:
            return True
        
        matches_rules = False
        
        for rule in self.rules:
            matches_rules = rule(
                request,
                ip_address
            )

        return matches_rules



from hedra.core.engines.types.common.url import URL
from typing import Dict, Optional, Coroutine, Callable

try:

    from opentelemetry.trace import Span

except ImportError:
    Span = object
    

class Request:
    url: URL
    method: str
    headers: Dict[str, str]


class Response:
    url: URL
    method: str
    headers: Dict[str, str]
    status: Optional[int]
    error: Exception


RequestHook = Optional[
    Callable[
        [
            Span, 
            Request
        ], 
        None
    ]
]


ResponseHook = Optional[
    Callable[
        [
            Span,
            Response,
        ],
        None,
    ]
]


TraceSignal = Callable[
    [   
        Span,
        Request,
        Response

    ], 
    Coroutine[
        None, 
        None, 
        None
    ]
]


UrlFilter = Callable[
    [
        str
    ], 
    str
]



import asyncio
from hedra.distributed.models.http import Limit
from .base_limiter import BaseLimiter


class LeakyBucketLimiter(BaseLimiter):

    __slots__ = (
        "max_rate",
        "time_period",
        "_rate_per_sec",
        "_level",
        "_waiters",
        "_loop",
        "_last_check"
    )

    def __init__(
        self,
        limit: Limit
     ) -> None:
        super().__init__(
            limit.max_requests,
            limit.period,
            reject_requests=limit.reject_requests
        )

        self._level = 0.0
        self._last_check = 0.0

    def _leak(self) -> None:

        if self._level:
            elapsed = self._loop.time() - self._last_check
            decrement = elapsed * self._rate_per_sec
            self._level = max(self._level - decrement, 0)            

        self._last_check = self._loop.time()

    def has_capacity(self, amount: float = 1) -> bool:

        self._leak()
        requested = self._level + amount

        if requested < self.max_rate:

            for fut in self._waiters.values():
                if not fut.done():
                    fut.set_result(True)
                    break

        return requested <= self.max_rate



import math
import asyncio
import random
from hedra.distributed.env import Env, load_env
from hedra.distributed.env.time_parser import TimeParser
from hedra.distributed.middleware.base import (
    Middleware,
    MiddlewareType
)
from hedra.distributed.middleware.base.types import RequestHandler
from hedra.distributed.models.http import (
    Request,
    Response
)
from hedra.distributed.rate_limiting.limiters import SlidingWindowLimiter
from typing import (
    Optional,
    Union
)
from .circuit_breaker_state import CircuitBreakerState


class CircuitBreaker(Middleware):

    def __init__(
        self,
        failure_threshold: Optional[float]=None,
        failure_window: Optional[str]=None,
        handler_timeout: Optional[str]=None,
        rejection_sensitivity: Optional[float]=None
    ) -> None:
        super().__init__(
            self.__class__.__name__,
            middleware_type=MiddlewareType.CALL
        )

        env = load_env(Env)

        if failure_threshold is None:
            failure_threshold = env.MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_THRESHOLD
        
        if failure_window is None:
            failure_window = env.MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_FAILURE_WINDOW

        if handler_timeout is None:
            handler_timeout = env.MERCURY_SYNC_HTTP_HANDLER_TIMEOUT

        if rejection_sensitivity is None:
            rejection_sensitivity = env.MERCURY_SYNC_HTTP_CIRCUIT_BREAKER_REJECTION_SENSITIVITY

        self.failure_threshold = failure_threshold
        self.rejection_sensitivity = rejection_sensitivity

        self.failure_window = TimeParser(failure_window).time
        self.handler_timeout = TimeParser(handler_timeout).time
        self._limiter_failure_window = failure_window

        self.overload = 0
        self.failed = 0
        self.succeeded = 0
        self.total_completed = 0

        self._rate_per_sec = 0
        self._rate_per_sec_succeeded = 0
        self._rate_per_sec_failed = 0

        self._previous_count = 0
        self._previous_count_succeeded = 0
        self._previous_count_failed = 0

        self.wraps: bool = False

        self._loop: Union[asyncio.AbstractEventLoop, None] = None
        self._current_time: Union[float, None]=None
        self._breaker_state = CircuitBreakerState.CLOSED

        self._limiter: Union[SlidingWindowLimiter, None]=None

        self._closed_window_start: Union[float, None] = None
        self._closed_elapsed = 0

        self._half_open_window_start: Union[float, None] = None
        self._half_open_elapsed = 0

    def trip_breaker(self) -> bool:

        failed_rate_threshold = max(
            self._rate_per_sec * self.failure_threshold,
            1
        )

        return int(self._rate_per_sec_failed) > int(failed_rate_threshold)
    
    def reject_request(self) -> bool:

        if (self._loop.time() - self._current_time) > self.failure_window:
            self._current_time = math.floor(
                self._loop.time()/self.failure_window
            ) * self.failure_window

            self._previous_count = self.total_completed
            self._previous_count_succeeded = self.succeeded
            self._previous_count_failed = self.failed

            self.failed = 0
            self.succeeded = 0
            self.total_completed = 0

        self._rate_per_sec = (
            self._previous_count * (
                self.failure_window - (self._loop.time() - self._current_time)
            )/self.failure_window
        ) + self.total_completed


        self._rate_per_sec_succeeded = (
            self._previous_count_succeeded * (
                self.failure_window - (self._loop.time() - self._current_time)
            )/self.failure_window
        ) + self.succeeded

        self._rate_per_sec_failed = (
            self._previous_count_failed * (
                self.failure_window - (self._loop.time() - self._current_time)
            )/self.failure_window
        ) + self.failed

        success_rate = self._rate_per_sec_succeeded/(1 - self.failure_threshold)

        rejection_probability = max(
            (self._rate_per_sec - success_rate)/(self._rate_per_sec + 1),
            0
        )**(1/self.rejection_sensitivity)
        
        return random.random() < rejection_probability

    async def __setup__(self):
        self._loop = asyncio.get_event_loop()
        self._current_time = self._loop.time()

    async def __run__(
        self,
        request: Request,
        handler: RequestHandler
    ):
        
        reject = self.reject_request()

        if self._breaker_state == CircuitBreakerState.OPEN and self._closed_elapsed < self.failure_window:
            self._closed_elapsed = self._loop.time() - self._closed_window_start
            reject = True

        elif self._breaker_state == CircuitBreakerState.OPEN:

            self._breaker_state = CircuitBreakerState.HALF_OPEN

            self._half_open_window_start = self._loop.time()
            self._closed_elapsed = 0

        if self._breaker_state == CircuitBreakerState.HALF_OPEN and self._half_open_elapsed < self.failure_window:
            self._half_open_elapsed = self._loop.time() - self._half_open_window_start

        elif self._breaker_state == CircuitBreakerState.HALF_OPEN:
            self._breaker_state = CircuitBreakerState.CLOSED
            self._half_open_elapsed = 0

        if reject:
            response = Response(
                request.path,
                request.method,
                headers={
                    'x-mercury-sync-overload': True
                }
            )

            status = 503
        
        else:

            try:

                response, status = await asyncio.wait_for(
                    handler(request),
                    timeout=self.handler_timeout
                )


                if self.wraps is False:
                    response = Response(
                        request.path,
                        request.method,
                        headers=handler.response_headers,
                        data=response
                    )

                
            except Exception:

                response = Response(
                    request.path,
                    request.method
                )

                status = 504
            
            # Don't count rejections toward failure stats.
            if status >= 400:
                self.failed += 1

            elif status < 400:
                self.succeeded += 1
            
            self.total_completed += 1

        breaker_open = self._breaker_state == CircuitBreakerState.CLOSED or self._breaker_state == CircuitBreakerState.HALF_OPEN

        if self.trip_breaker() and breaker_open:

            self._breaker_state = CircuitBreakerState.OPEN
            reject = True

            self._closed_window_start = self._loop.time()
            self._half_open_elapsed = 0

        return (
            request,
            response,
            status
        )
            
            
            


        

        



import uuid
import json
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.playwright import (
    PlaywrightCommand,
    Page,
    Input,
    URL,
    Options,
    PlaywrightResult
)
from hedra.core.engines.types.common.types import RequestTypes
from hedra.data.parsers.parser_types.common.base_parser import BaseParser
from hedra.data.parsers.parser_types.common.result_validator import ResultValidator
from hedra.data.parsers.parser_types.common.parsing import (
    normalize_headers,
    parse_tags
)
from typing import Any, Coroutine, Dict
from .playwright_action_validator import (
    PlaywrightActionValidator,
    PlaywrightInputValidator,
    PlaywrightOptionsValidator,
    PlaywrightPageValidator,
    PlaywrightURLValidator
)



class PlaywrightResultParser(BaseParser):

    def __init__(
        self,
        config: Config,
        options: Dict[str, Any]={}
    ) -> None:
        super().__init__(
            PlaywrightResultParser.__name__,
            config,
            RequestTypes.UDP,
            options
        )

    async def parse(
        self, 
        result_data: Dict[str, Any]
    ) -> Coroutine[Any, Any, Coroutine[Any, Any, PlaywrightResult]]:
        
        normalized_headers = normalize_headers(result_data)
        tags_data = parse_tags(result_data)

        playwright_input_args = result_data.get('args')
        if isinstance(playwright_input_args, (str, bytes, bytearray,)):
            playwright_input_args = playwright_input_args.split(',')

        playwright_options_extra = result_data.get('extra')
        if isinstance(playwright_options_extra, (str, bytes, bytearray,)):
            playwright_options_extra = json.loads(playwright_options_extra)

        generator_action = PlaywrightActionValidator(
            name=result_data.get('name'),
            command=result_data.get('command'),
            page=PlaywrightPageValidator(
                selector=result_data.get('selector'),
                attribute=result_data.get('attribute'),
                x_coordinate=result_data.get('x_coordinate'),
                y_coordinate=result_data.get('y_coordinate'),
                frame=result_data.get('frame')
            ),
            url=PlaywrightURLValidator(
                location=result_data.get('location'),
                headers=normalized_headers
            ),
            input=PlaywrightInputValidator(
                key=result_data.get('key'),
                text=result_data.get('text'),
                expression=result_data.get('expression'),
                args=playwright_input_args,
                filepath=result_data.get('filepath'),
                file=result_data.get('file'),
                path=result_data.get('path'),
                option=result_data.get('option'),
                by_label=result_data.get('by_label', False),
                by_value=result_data.get('by_value', False)
            ),
            options=PlaywrightOptionsValidator(
                event=result_data.get('event'),
                option=result_data.get('option'),
                is_checked=result_data.get('is_checked', False),
                timeout=result_data.get('timeout'),
                extra=playwright_options_extra,
                switch_by=result_data.get('switch_by')
            ),
            weight=result_data.get('weight'),
            order=result_data.get('order'),
            user=result_data.get('user'),
            tags=tags_data
        )

        action = PlaywrightCommand(
            generator_action.name,
            generator_action.command,
            url=URL(**generator_action.url.dict(
                exclude_none=True
            )),
            page=Page(**generator_action.page.dict(
                exclude_none=True
            )),
            input=Input(**generator_action.input.dict(
                exclude_none=True
            )),
            options=Options(**generator_action.options.dict(
                exclude_none=True
            )),
            user=generator_action.user,
            tags=[
                tag.dict() for tag in generator_action.tags
            ]
        )

        body = result_data.get('body')
        if isinstance(body, str):
            body = body.encode()

        result_validator = ResultValidator(
            error=result_data.get('error'), 
            status=result_data.get('status'),
            reason=result_data.get('reason'),
            params=result_data.get('params'),
            wait_start=result_data.get('wait_start'),
            start=result_data.get('start'),
            connect_end=result_data.get('connect_end'),
            write_end=result_data.get('write_end'),
            complete=result_data.get('complete'),
            checks=result_data.get('checks')
        )

        result = PlaywrightResult(
            action,
            error=Exception(result_validator.error) if result_validator.error else None
        )

        result.checks = result_validator.checks
        result.wait_start = result_validator.wait_start
        result.start = result_validator.start
        result.connect_end = result_validator.connect_end
        result.write_end = result_validator.write_end
        result.complete = result_validator.complete
        
        result.type= RequestTypes.PLAYWRIGHT

        return result



